from functools import partial
from typing import Callable, Dict

from .image import pil_exporter
from .landmark import ljson_exporter, pts_exporter
from .pickle import pickle_exporter
from .video import ffmpeg_video_exporter

landmark_types: Dict[str, Callable] = {".ljson": ljson_exporter, ".pts": pts_exporter}


image_types = {
    ".bmp": pil_exporter,
    ".dib": pil_exporter,
    ".dcx": pil_exporter,
    ".eps": pil_exporter,
    ".ps": pil_exporter,
    ".gif": pil_exporter,
    ".im": pil_exporter,
    ".jpg": pil_exporter,
    ".jpe": pil_exporter,
    ".jpeg": pil_exporter,
    ".pcd": pil_exporter,
    ".pcx": pil_exporter,
    ".png": pil_exporter,
    ".pbm": pil_exporter,
    ".pgm": pil_exporter,
    ".ppm": pil_exporter,
    ".psd": pil_exporter,
    ".tif": pil_exporter,
    ".tiff": pil_exporter,
    ".xbm": pil_exporter,
    ".xpm": pil_exporter,
}


pickle_types = {
    ".pkl": pickle_exporter,
    ".pkl.gz": pickle_exporter,
}


video_types = {
    ".mov": ffmpeg_video_exporter,
    ".avi": ffmpeg_video_exporter,
    ".mpg": ffmpeg_video_exporter,
    ".mpeg": ffmpeg_video_exporter,
    ".mp4": ffmpeg_video_exporter,
    ".mkv": ffmpeg_video_exporter,
    ".wmv": ffmpeg_video_exporter,
    ".gif": partial(ffmpeg_video_exporter, codec=None),
}



from collections import OrderedDict
import numpy as np

from ..base import labeller_func, validate_input, connectivity_from_array


@labeller_func(group_label="face_bu3dfe_83")
def face_bu3dfe_83_to_face_bu3dfe_83(pcloud):
    r"""
    Apply the BU-3DFE (Binghamton University 3D Facial Expression)
    Database 83-point facial semantic labels.

    The semantic labels applied are as follows:

      - right_eye
      - left_eye
      - right_eyebrow
      - left_eyebrow
      - right_nose
      - left_nose
      - nostrils
      - outer_mouth
      - inner_mouth
      - jaw

    References
    ----------
    .. [1] http://www.cs.binghamton.edu/~lijun/Research/3DFE/3DFE_Analysis.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 83
    validate_input(pcloud, n_expected_points)

    reye_indices = np.arange(0, 8)
    leye_indices = np.arange(8, 16)
    rbrow_indices = np.arange(16, 26)
    lbrow_indices = np.arange(26, 36)
    rnose_indices = np.arange(36, 39)
    lnose_indices = np.arange(39, 42)
    nostril_indices = np.arange(42, 48)
    outermouth_indices = np.arange(48, 60)
    innermouth_indices = np.arange(60, 68)
    jaw_indices = np.arange(68, 83)

    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    rbrow_connectivity = connectivity_from_array(rbrow_indices, close_loop=True)
    lbrow_connectivity = connectivity_from_array(lbrow_indices, close_loop=True)
    rnose_connectivity = connectivity_from_array(rnose_indices)
    nostril_connectivity = connectivity_from_array(nostril_indices)
    lnose_connectivity = connectivity_from_array(lnose_indices)
    outermouth_connectivity = connectivity_from_array(
        outermouth_indices, close_loop=True
    )
    innermouth_connectivity = connectivity_from_array(
        innermouth_indices, close_loop=True
    )
    jaw_connectivity = connectivity_from_array(jaw_indices)

    all_connectivity = np.vstack(
        [
            reye_connectivity,
            leye_connectivity,
            rbrow_connectivity,
            lbrow_connectivity,
            rnose_connectivity,
            nostril_connectivity,
            lnose_connectivity,
            outermouth_connectivity,
            innermouth_connectivity,
            jaw_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["right_eye"] = reye_indices
    mapping["left_eye"] = leye_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_nose"] = rnose_indices
    mapping["left_nose"] = lnose_indices
    mapping["nostrils"] = nostril_indices
    mapping["outer_mouth"] = outermouth_indices
    mapping["inner_mouth"] = innermouth_indices
    mapping["jaw"] = jaw_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping



import sys
from pathlib import Path

import black
import pytest
from click.testing import CliRunner

menpo_root = Path(__file__).parent.parent


@pytest.mark.skipif(
    sys.version_info[:2] != (3, 7),
    reason="Only run on one Python version to save testing time",
)
def test_black():
    runner = CliRunner()
    # Manually ignore the file generated by miniver
    result = runner.invoke(
        black.main, [str(menpo_root), "--check", "--exclude", "_static_version.py"]
    )
    assert result.exit_code == 0, result.output



import numpy as np

from menpo.visualize.base import Renderer

# The colour map used for all lines and markers
GLOBAL_CMAP = "jet"


class MatplotlibRenderer(Renderer):
    r"""
    Abstract class for rendering visualizations using Matplotlib.

    Parameters
    ----------
    figure_id : `int` or ``None``
        A figure id or ``None``. ``None`` assumes we maintain the Matplotlib
        state machine and use `plt.gcf()`.
    new_figure : `bool`
        If ``True``, it creates a new figure to render on.
    """

    def __init__(self, figure_id, new_figure):
        super(MatplotlibRenderer, self).__init__(figure_id, new_figure)

        # Set up data for saving
        self._supported_ext = self.figure.canvas.get_supported_filetypes().keys()
        # Create the extensions map, have to add . in front of the extensions
        # and map every extension to the savefig method
        n_ext = len(self._supported_ext)
        func_list = [lambda obj, fp, **kwargs: self.figure.savefig(fp, **obj)] * n_ext
        self._extensions_map = dict(
            zip(["." + s for s in self._supported_ext], func_list)
        )

    def get_figure(self):
        r"""
        Gets the figure specified by the combination of ``self.figure_id`` and
        ``self.new_figure``. If ``self.figure_id == None`` then ``plt.gcf()``
        is used. ``self.figure_id`` is also set to the correct id of the figure
        if a new figure is created.

        Returns
        -------
        figure : Matplotlib figure object
            The figure we will be rendering on.
        """
        import matplotlib.pyplot as plt

        if self.new_figure or self.figure_id is not None:
            self.figure = plt.figure(self.figure_id)
        else:
            self.figure = plt.gcf()

        self.figure_id = self.figure.number

        return self.figure

    def save_figure(
        self,
        filename,
        format="png",
        dpi=None,
        face_colour="w",
        edge_colour="w",
        orientation="portrait",
        paper_type="letter",
        transparent=False,
        pad_inches=0.1,
        overwrite=False,
    ):
        r"""
        Method for saving the figure of the current `figure_id` to file.

        Parameters
        ----------
        filename : `str` or `file`-like object
            The string path or file-like object to save the figure at/into.
        format : `str`
            The format to use. This must match the file path if the file path is
            a `str`.
        dpi : `int` > 0 or ``None``, optional
            The resolution in dots per inch.
        face_colour : See Below, optional
            The face colour of the figure rectangle.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or
                ``(3, )`` `ndarray`
                or
                `list` of len 3

        edge_colour : See Below, optional
            The edge colour of the figure rectangle.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or
                ``(3, )`` `ndarray`
                or
                `list` of len 3

        orientation : {``portrait``, ``landscape``}, optional
            The page orientation.
        paper_type : See Below, optional
            The type of the paper.
            Example options ::

                {``letter``, ``legal``, ``executive``, ``ledger``,
                 ``a0`` through ``a10``, ``b0` through ``b10``}

        transparent : `bool`, optional
            If ``True``, the axes patches will all be transparent; the figure
            patch will also be transparent unless `face_colour` and/or
            `edge_colour` are specified. This is useful, for example, for
            displaying a plot on top of a coloured background on a web page.
            The transparency of these patches will be restored to their original
            values upon exit of this function.
        pad_inches : `float`, optional
            Amount of padding around the figure.
        overwrite : `bool`, optional
            If ``True``, the file will be overwritten if it already exists.
        """
        from menpo.io.output.base import _export

        save_fig_args = {
            "dpi": dpi,
            "facecolour": face_colour,
            "edgecolour": edge_colour,
            "orientation": orientation,
            "papertype": paper_type,
            "format": format,
            "transparent": transparent,
            "pad_inches": pad_inches,
            "bbox_inches": "tight",
            "frameon": None,
        }
        # Use the export code so that we have a consistent interface
        _export(
            save_fig_args, filename, self._extensions_map, format, overwrite=overwrite
        )

    def clear_figure(self):
        r"""
        Method for clearing the current figure.
        """
        self.figure.clf()

    def force_draw(self):
        r"""
        Method for forcing the current figure to render.
        """
        import matplotlib.pyplot as plt

        plt.show()


class MatplotlibSubplots(object):
    def _subplot_layout(self, num_subplots):
        if num_subplots < 2:
            return [1, 1]
        while self._is_prime(num_subplots) and num_subplots > 4:
            num_subplots += 1
        p = self._factor(num_subplots)
        if len(p) == 1:
            p.insert(0, 1)
            return p
        while len(p) > 2:
            if len(p) >= 4:
                p[0] = p[0] * p[-2]
                p[1] = p[1] * p[-1]
                del p[-2:]
            else:
                p[0] = p[0] * p[1]
                del p[1]
            p.sort()
            # Reformat if the column/row ratio is too large: we want a roughly
        # square design
        while (p[1] / p[0]) > 2.5:
            p = self._subplot_layout(num_subplots + 1)
        return p

    def _factor(self, n):
        gaps = [1, 2, 2, 4, 2, 4, 2, 4, 6, 2, 6]
        length, cycle = 11, 3
        f, fs, next_ind = 2, [], 0
        while f * f <= n:
            while n % f == 0:
                fs.append(f)
                n /= f
            f += gaps[next_ind]
            next_ind += 1
            if next_ind == length:
                next_ind = cycle
        if n > 1:
            fs.append(n)
        return fs

    def _is_prime(self, n):
        if n == 2 or n == 3:
            return True
        if n < 2 or n % 2 == 0:
            return False
        if n < 9:
            return True
        if n % 3 == 0:
            return False
        r = int(n ** 0.5)
        f = 5
        while f <= r:
            if n % f == 0:
                return False
            if n % (f + 2) == 0:
                return False
            f += 6
        return True


def _parse_cmap(cmap_name=None, image_shape_len=3):
    import matplotlib.cm as cm

    if cmap_name is not None:
        return cm.get_cmap(cmap_name)
    else:
        if image_shape_len == 2:
            # Single channels are viewed in Gray by default
            return cm.gray
        else:
            return None


def _parse_axes_limits(min_x, max_x, min_y, max_y, axes_x_limits, axes_y_limits):
    if isinstance(axes_x_limits, int):
        axes_x_limits = float(axes_x_limits)
    if isinstance(axes_y_limits, int):
        axes_y_limits = float(axes_y_limits)
    if isinstance(axes_x_limits, float):
        pad = (max_x - min_x) * axes_x_limits
        axes_x_limits = [min_x - pad, max_x + pad]
    if isinstance(axes_y_limits, float):
        pad = (max_y - min_y) * axes_y_limits
        axes_y_limits = [min_y - pad, max_y + pad]
    return axes_x_limits, axes_y_limits


def _set_axes_options(
    ax,
    render_axes=True,
    inverted_y_axis=False,
    axes_font_name="sans-serif",
    axes_font_size=10,
    axes_font_style="normal",
    axes_font_weight="normal",
    axes_x_limits=None,
    axes_y_limits=None,
    axes_x_ticks=None,
    axes_y_ticks=None,
    axes_x_label=None,
    axes_y_label=None,
    title=None,
):
    if render_axes:
        # render axes
        ax.set_axis_on()
        # set font options
        for l in ax.get_xticklabels() + ax.get_yticklabels():
            l.set_fontsize(axes_font_size)
            l.set_fontname(axes_font_name)
            l.set_fontstyle(axes_font_style)
            l.set_fontweight(axes_font_weight)
        # set ticks
        if axes_x_ticks is not None:
            ax.set_xticks(axes_x_ticks)
        if axes_y_ticks is not None:
            ax.set_yticks(axes_y_ticks)
        # set labels and title
        if axes_x_label is None:
            axes_x_label = ""
        if axes_y_label is None:
            axes_y_label = ""
        if title is None:
            title = ""
        ax.set_xlabel(
            axes_x_label,
            fontsize=axes_font_size,
            fontname=axes_font_name,
            fontstyle=axes_font_style,
            fontweight=axes_font_weight,
        )
        ax.set_ylabel(
            axes_y_label,
            fontsize=axes_font_size,
            fontname=axes_font_name,
            fontstyle=axes_font_style,
            fontweight=axes_font_weight,
        )
        ax.set_title(
            title,
            fontsize=axes_font_size,
            fontname=axes_font_name,
            fontstyle=axes_font_style,
            fontweight=axes_font_weight,
        )
    else:
        # do not render axes
        ax.set_axis_off()
        # also remove the ticks to get rid of the white area
        ax.set_xticks([])
        ax.set_yticks([])

    # set axes limits
    if axes_x_limits is not None:
        ax.set_xlim(np.sort(axes_x_limits))
    if axes_y_limits is None:
        axes_y_limits = ax.get_ylim()
    if inverted_y_axis:
        ax.set_ylim(np.sort(axes_y_limits)[::-1])
    else:
        ax.set_ylim(np.sort(axes_y_limits))


def _set_grid_options(render_grid=True, grid_line_style="--", grid_line_width=2):
    import matplotlib.pyplot as plt

    if render_grid:
        plt.grid("on", linestyle=grid_line_style, linewidth=grid_line_width)
    else:
        plt.grid("off")


def _set_figure_size(fig, figure_size=(7, 7)):
    if figure_size is not None:
        fig.set_size_inches(np.asarray(figure_size))


def _set_numbering(
    ax,
    centers,
    render_numbering=True,
    numbers_horizontal_align="center",
    numbers_vertical_align="bottom",
    numbers_font_name="sans-serif",
    numbers_font_size=10,
    numbers_font_style="normal",
    numbers_font_weight="normal",
    numbers_font_colour="k",
):
    if render_numbering:
        for k, p in enumerate(centers):
            ax.annotate(
                str(k),
                xy=(p[0], p[1]),
                horizontalalignment=numbers_horizontal_align,
                verticalalignment=numbers_vertical_align,
                size=numbers_font_size,
                family=numbers_font_name,
                fontstyle=numbers_font_style,
                fontweight=numbers_font_weight,
                color=numbers_font_colour,
            )


def _set_legend(
    ax,
    legend_handles,
    render_legend=True,
    legend_title="",
    legend_font_name="sans-serif",
    legend_font_style="normal",
    legend_font_size=10,
    legend_font_weight="normal",
    legend_marker_scale=None,
    legend_location=2,
    legend_bbox_to_anchor=(1.05, 1.0),
    legend_border_axes_pad=None,
    legend_n_columns=1,
    legend_horizontal_spacing=None,
    legend_vertical_spacing=None,
    legend_border=True,
    legend_border_padding=None,
    legend_shadow=False,
    legend_rounded_corners=False,
):
    if render_legend:
        # Options related to legend's font
        prop = {
            "family": legend_font_name,
            "size": legend_font_size,
            "style": legend_font_style,
            "weight": legend_font_weight,
        }

        # Render legend
        ax.legend(
            handles=legend_handles,
            title=legend_title,
            prop=prop,
            loc=legend_location,
            bbox_to_anchor=legend_bbox_to_anchor,
            borderaxespad=legend_border_axes_pad,
            ncol=legend_n_columns,
            columnspacing=legend_horizontal_spacing,
            labelspacing=legend_vertical_spacing,
            frameon=legend_border,
            borderpad=legend_border_padding,
            shadow=legend_shadow,
            fancybox=legend_rounded_corners,
            markerscale=legend_marker_scale,
        )


class MatplotlibImageViewer2d(MatplotlibRenderer):
    def __init__(self, figure_id, new_figure, image):
        super(MatplotlibImageViewer2d, self).__init__(figure_id, new_figure)
        self.image = image
        self.axes_list = []

    def render(
        self,
        interpolation="bilinear",
        cmap_name=None,
        alpha=1.0,
        render_axes=False,
        axes_font_name="sans-serif",
        axes_font_size=10,
        axes_font_style="normal",
        axes_font_weight="normal",
        axes_x_limits=None,
        axes_y_limits=None,
        axes_x_ticks=None,
        axes_y_ticks=None,
        figure_size=(7, 7),
    ):
        import matplotlib.pyplot as plt

        # parse colour map argument
        cmap = _parse_cmap(cmap_name=cmap_name, image_shape_len=len(self.image.shape))

        # parse axes limits
        axes_x_limits, axes_y_limits = _parse_axes_limits(
            0.0,
            self.image.shape[1],
            0.0,
            self.image.shape[0],
            axes_x_limits,
            axes_y_limits,
        )

        # render image
        plt.imshow(self.image, cmap=cmap, interpolation=interpolation, alpha=alpha)

        # store axes object
        ax = plt.gca()
        self.axes_list = [ax]

        # set axes options
        _set_axes_options(
            ax,
            render_axes=render_axes,
            inverted_y_axis=True,
            axes_font_name=axes_font_name,
            axes_font_size=axes_font_size,
            axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits,
            axes_x_ticks=axes_x_ticks,
            axes_y_ticks=axes_y_ticks,
        )

        # set figure size
        _set_figure_size(self.figure, figure_size)

        return self


class MatplotlibImageSubplotsViewer2d(MatplotlibRenderer, MatplotlibSubplots):
    def __init__(self, figure_id, new_figure, image):
        super(MatplotlibImageSubplotsViewer2d, self).__init__(figure_id, new_figure)
        self.image = image
        self.num_subplots = self.image.shape[2]
        self.plot_layout = self._subplot_layout(self.num_subplots)
        self.axes_list = []

    def render(
        self,
        interpolation="bilinear",
        cmap_name=None,
        alpha=1.0,
        render_axes=False,
        axes_font_name="sans-serif",
        axes_font_size=10,
        axes_font_style="normal",
        axes_font_weight="normal",
        axes_x_limits=None,
        axes_y_limits=None,
        axes_x_ticks=None,
        axes_y_ticks=None,
        figure_size=(7, 7),
    ):
        import matplotlib.pyplot as plt

        # parse colour map argument
        cmap = _parse_cmap(cmap_name=cmap_name, image_shape_len=2)

        # parse axes limits
        axes_x_limits, axes_y_limits = _parse_axes_limits(
            0.0,
            self.image.shape[1],
            0.0,
            self.image.shape[0],
            axes_x_limits,
            axes_y_limits,
        )

        p = self.plot_layout
        for i in range(self.image.shape[2]):
            # create subplot and append the axes object
            ax = plt.subplot(p[0], p[1], 1 + i)
            self.axes_list.append(ax)

            # render image
            plt.imshow(
                self.image[:, :, i], cmap=cmap, interpolation=interpolation, alpha=alpha
            )

            # set axes options
            _set_axes_options(
                ax,
                render_axes=render_axes,
                inverted_y_axis=True,
                axes_font_name=axes_font_name,
                axes_font_size=axes_font_size,
                axes_font_style=axes_font_style,
                axes_font_weight=axes_font_weight,
                axes_x_limits=axes_x_limits,
                axes_y_limits=axes_y_limits,
                axes_x_ticks=axes_x_ticks,
                axes_y_ticks=axes_y_ticks,
            )

        # set figure size
        _set_figure_size(self.figure, figure_size)

        return self


class MatplotlibPointGraphViewer2d(MatplotlibRenderer):
    def __init__(self, figure_id, new_figure, points, edges):
        super(MatplotlibPointGraphViewer2d, self).__init__(figure_id, new_figure)
        self.points = points
        self.edges = edges

    def render(
        self,
        image_view=False,
        render_lines=True,
        line_colour="r",
        line_style="-",
        line_width=1,
        render_markers=True,
        marker_style="o",
        marker_size=5,
        marker_face_colour="r",
        marker_edge_colour="k",
        marker_edge_width=1.0,
        render_numbering=False,
        numbers_horizontal_align="center",
        numbers_vertical_align="bottom",
        numbers_font_name="sans-serif",
        numbers_font_size=10,
        numbers_font_style="normal",
        numbers_font_weight="normal",
        numbers_font_colour="k",
        render_axes=True,
        axes_font_name="sans-serif",
        axes_font_size=10,
        axes_font_style="normal",
        axes_font_weight="normal",
        axes_x_limits=None,
        axes_y_limits=None,
        axes_x_ticks=None,
        axes_y_ticks=None,
        figure_size=(7, 7),
        label=None,
    ):
        from matplotlib import collections as mc
        import matplotlib.pyplot as plt

        # Flip x and y for viewing if points are tied to an image
        points = self.points[:, ::-1] if image_view else self.points

        # parse axes limits
        min_x, min_y = np.min(points, axis=0)
        max_x, max_y = np.max(points, axis=0)
        axes_x_limits, axes_y_limits = _parse_axes_limits(
            min_x, max_x, min_y, max_y, axes_x_limits, axes_y_limits
        )

        # get current axes object
        ax = plt.gca()

        # Check if graph has edges to be rendered (for example a PointCloud
        # won't have any edges)
        if render_lines and np.array(self.edges).shape[0] > 0:
            # Get edges to be rendered
            lines = zip(points[self.edges[:, 0], :], points[self.edges[:, 1], :])

            # Draw line objects
            lc = mc.LineCollection(
                lines,
                colors=line_colour,
                linestyles=line_style,
                linewidths=line_width,
                cmap=GLOBAL_CMAP,
                label=label,
            )
            ax.add_collection(lc)

            # If a label is defined, it should only be applied to the lines, of
            # a PointGraph, which represent each one of the labels, unless a
            # PointCloud is passed in.
            label = None
            ax.autoscale()

        if render_markers:
            plt.plot(
                points[:, 0],
                points[:, 1],
                linewidth=0,
                markersize=marker_size,
                marker=marker_style,
                markeredgewidth=marker_edge_width,
                markeredgecolor=marker_edge_colour,
                markerfacecolor=marker_face_colour,
                label=label,
            )

        # set numbering
        _set_numbering(
            ax,
            points,
            render_numbering=render_numbering,
            numbers_horizontal_align=numbers_horizontal_align,
            numbers_vertical_align=numbers_vertical_align,
            numbers_font_name=numbers_font_name,
            numbers_font_size=numbers_font_size,
            numbers_font_style=numbers_font_style,
            numbers_font_weight=numbers_font_weight,
            numbers_font_colour=numbers_font_colour,
        )

        # set axes options
        _set_axes_options(
            ax,
            render_axes=render_axes,
            inverted_y_axis=image_view,
            axes_font_name=axes_font_name,
            axes_font_size=axes_font_size,
            axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits,
            axes_x_ticks=axes_x_ticks,
            axes_y_ticks=axes_y_ticks,
        )

        # set equal aspect ratio
        ax.set_aspect("equal", adjustable="box")

        # set figure size
        _set_figure_size(self.figure, figure_size)

        return self


class MatplotlibLandmarkViewer2d(MatplotlibRenderer):
    def __init__(self, figure_id, new_figure, group, landmark_group):
        super(MatplotlibLandmarkViewer2d, self).__init__(figure_id, new_figure)
        self.group = group
        self.landmark_group = landmark_group

    def render(
        self,
        image_view=False,
        render_lines=True,
        line_colour="r",
        line_style="-",
        line_width=1,
        render_markers=True,
        marker_style="o",
        marker_size=5,
        marker_face_colour="r",
        marker_edge_colour="k",
        marker_edge_width=1.0,
        render_numbering=False,
        numbers_horizontal_align="center",
        numbers_vertical_align="bottom",
        numbers_font_name="sans-serif",
        numbers_font_size=10,
        numbers_font_style="normal",
        numbers_font_weight="normal",
        numbers_font_colour="k",
        render_legend=True,
        legend_title="",
        legend_font_name="sans-serif",
        legend_font_style="normal",
        legend_font_size=10,
        legend_font_weight="normal",
        legend_marker_scale=None,
        legend_location=2,
        legend_bbox_to_anchor=(1.05, 1.0),
        legend_border_axes_pad=None,
        legend_n_columns=1,
        legend_horizontal_spacing=None,
        legend_vertical_spacing=None,
        legend_border=True,
        legend_border_padding=None,
        legend_shadow=False,
        legend_rounded_corners=False,
        render_axes=True,
        axes_font_name="sans-serif",
        axes_font_size=10,
        axes_font_style="normal",
        axes_font_weight="normal",
        axes_x_limits=None,
        axes_y_limits=None,
        axes_x_ticks=None,
        axes_y_ticks=None,
        figure_size=(7, 7),
    ):
        import matplotlib.lines as mlines
        import matplotlib.pyplot as plt

        # Regarding the labels colours, we may get passed either no colours (in
        # which case we generate random colours) or a single colour to colour
        # all the labels with
        # TODO: All marker and line options could be defined as lists...
        n_labels = self.landmark_group.n_labels
        line_colour = _check_colours_list(
            render_lines,
            line_colour,
            n_labels,
            "Must pass a list of line colours with length n_labels or a single "
            "line colour for all labels.",
        )
        marker_face_colour = _check_colours_list(
            render_markers,
            marker_face_colour,
            n_labels,
            "Must pass a list of marker face colours with length n_labels or "
            "a single marker face colour for all labels.",
        )
        marker_edge_colour = _check_colours_list(
            render_markers,
            marker_edge_colour,
            n_labels,
            "Must pass a list of marker edge colours with length n_labels or "
            "a single marker edge colour for all labels.",
        )

        # check axes limits
        points = self.landmark_group.points
        if image_view:
            min_y, min_x = np.min(points, axis=0)
            max_y, max_x = np.max(points, axis=0)
        else:
            min_x, min_y = np.min(points, axis=0)
            max_x, max_y = np.max(points, axis=0)
        axes_x_limits, axes_y_limits = _parse_axes_limits(
            min_x, max_x, min_y, max_y, axes_x_limits, axes_y_limits
        )

        # get pointcloud of each label
        sub_pointclouds = self._build_sub_pointclouds()

        # initialize legend_handles list
        legend_handles = []

        # for each pointcloud
        for i, (label, pc) in enumerate(sub_pointclouds):
            # render pointcloud
            pc.view(
                figure_id=self.figure_id,
                image_view=image_view,
                render_lines=render_lines,
                line_colour=line_colour[i],
                line_style=line_style,
                line_width=line_width,
                render_markers=render_markers,
                marker_style=marker_style,
                marker_size=marker_size,
                marker_face_colour=marker_face_colour[i],
                marker_edge_colour=marker_edge_colour[i],
                marker_edge_width=marker_edge_width,
                render_numbering=render_numbering,
                numbers_horizontal_align=numbers_horizontal_align,
                numbers_vertical_align=numbers_vertical_align,
                numbers_font_name=numbers_font_name,
                numbers_font_size=numbers_font_size,
                numbers_font_style=numbers_font_style,
                numbers_font_weight=numbers_font_weight,
                numbers_font_colour=numbers_font_colour,
                render_axes=render_axes,
                axes_font_name=axes_font_name,
                axes_font_size=axes_font_size,
                axes_font_style=axes_font_style,
                axes_font_weight=axes_font_weight,
                axes_x_limits=axes_x_limits,
                axes_y_limits=axes_y_limits,
                axes_x_ticks=axes_x_ticks,
                axes_y_ticks=axes_y_ticks,
                figure_size=None,
            )

            # set legend entry
            if render_legend:
                tmp_line = line_style
                tmp_marker = marker_style if render_markers else "None"
                legend_handles.append(
                    mlines.Line2D(
                        [],
                        [],
                        linewidth=line_width,
                        linestyle=tmp_line,
                        color=line_colour[i],
                        marker=tmp_marker,
                        markersize=marker_size ** 0.5,
                        markeredgewidth=marker_edge_width,
                        markeredgecolor=marker_edge_colour[i],
                        markerfacecolor=marker_face_colour[i],
                        label="{0}: {1}".format(self.group, label),
                    )
                )
        # set legend
        _set_legend(
            plt.gca(),
            legend_handles,
            render_legend=render_legend,
            legend_title=legend_title,
            legend_font_name=legend_font_name,
            legend_font_style=legend_font_style,
            legend_font_size=legend_font_size,
            legend_font_weight=legend_font_weight,
            legend_marker_scale=legend_marker_scale,
            legend_location=legend_location,
            legend_bbox_to_anchor=legend_bbox_to_anchor,
            legend_border_axes_pad=legend_border_axes_pad,
            legend_n_columns=legend_n_columns,
            legend_horizontal_spacing=legend_horizontal_spacing,
            legend_vertical_spacing=legend_vertical_spacing,
            legend_border=legend_border,
            legend_border_padding=legend_border_padding,
            legend_shadow=legend_shadow,
            legend_rounded_corners=legend_rounded_corners,
        )

        # set figure size
        _set_figure_size(self.figure, figure_size)
        # If no limits are set then ensure that all sub-pointclouds fit in the
        # view
        if axes_x_limits is None and axes_y_limits is None:
            plt.autoscale()

        return self

    def _build_sub_pointclouds(self):
        return [
            (label, self.landmark_group.get_label(label))
            for label in self.landmark_group.labels
        ]


class MatplotlibAlignmentViewer2d(MatplotlibRenderer):
    def __init__(self, figure_id, new_figure, alignment_transform):
        super(MatplotlibAlignmentViewer2d, self).__init__(figure_id, new_figure)
        self.alignment_transform = alignment_transform

    def render(self, image=False, **kwargs):
        r"""
        Visualize how points are affected by the warp in 2 dimensions.
        """
        import matplotlib.pyplot as plt

        source = self.alignment_transform.source.points
        target = self.alignment_transform.target.points
        # a factor by which the minimum and maximum x and y values of the warp
        # will be increased by.
        x_margin_factor, y_margin_factor = 0.5, 0.5
        # the number of x and y samples to take
        n_x, n_y = 50, 50
        # {x y}_{min max} is the actual bounds on either source or target
        # landmarks
        x_min, y_min = np.vstack([target.min(0), source.min(0)]).min(0)
        x_max, y_max = np.vstack([target.max(0), source.max(0)]).max(0)
        x_margin = x_margin_factor * (x_max - x_min)
        y_margin = y_margin_factor * (y_max - y_min)
        # {x y}_{min max}_m is the bound once it has been grown by the factor
        # of the spread in that dimension
        x_min_m = x_min - x_margin
        x_max_m = x_max + x_margin
        y_min_m = y_min - y_margin
        y_max_m = y_max + y_margin
        # build sample points for the selected region
        x = np.linspace(x_min_m, x_max_m, n_x)
        y = np.linspace(y_min_m, y_max_m, n_y)
        xx, yy = np.meshgrid(x, y)
        sample_points = np.concatenate(
            [xx.reshape([-1, 1]), yy.reshape([-1, 1])], axis=1
        )
        warped_points = self.alignment_transform.apply(sample_points)
        delta = warped_points - sample_points
        # plot the sample points result
        x, y, = (
            0,
            1,
        )
        if image:
            # if we are overlaying points onto an image,
            # we have to account for the fact that axis 0 is typically
            # called 'y' and axis 1 is typically called 'x'. Flip them here
            x, y = y, x
        plt.quiver(sample_points[:, x], sample_points[:, y], delta[:, x], delta[:, y])
        delta = target - source
        # plot how the landmarks move from source to target
        plt.quiver(
            source[:, x],
            source[:, y],
            delta[:, x],
            delta[:, y],
            angles="xy",
            scale_units="xy",
            scale=1,
        )
        # rescale to the bounds
        plt.xlim((x_min_m, x_max_m))
        plt.ylim((y_min_m, y_max_m))
        if image:
            # if we are overlaying points on an image, axis0 (the 'y' axis)
            # is flipped.
            plt.gca().invert_yaxis()
        return self


class MatplotlibGraphPlotter(MatplotlibRenderer):
    def __init__(
        self,
        figure_id,
        new_figure,
        x_axis,
        y_axis,
        title=None,
        legend_entries=None,
        x_label=None,
        y_label=None,
        x_axis_limits=None,
        y_axis_limits=None,
        x_axis_ticks=None,
        y_axis_ticks=None,
    ):
        super(MatplotlibGraphPlotter, self).__init__(figure_id, new_figure)
        self.x_axis = x_axis
        self.y_axis = y_axis
        if legend_entries is None:
            legend_entries = ["Curve {}".format(i) for i in range(len(y_axis))]
        self.legend_entries = legend_entries
        self.title = title
        self.x_label = x_label
        self.y_label = y_label
        self.x_axis_ticks = x_axis_ticks
        self.y_axis_ticks = y_axis_ticks
        # parse axes limits
        min_x = np.min(x_axis)
        max_x = np.max(x_axis)
        min_y = np.min([np.min(l) for l in y_axis])
        max_y = np.max([np.max(l) for l in y_axis])
        self.x_axis_limits, self.y_axis_limits = _parse_axes_limits(
            min_x, max_x, min_y, max_y, x_axis_limits, y_axis_limits
        )

    def render(
        self,
        render_lines=True,
        line_colour="r",
        line_style="-",
        line_width=1,
        render_markers=True,
        marker_style="o",
        marker_size=6,
        marker_face_colour="r",
        marker_edge_colour="k",
        marker_edge_width=1.0,
        render_legend=True,
        legend_title="",
        legend_font_name="sans-serif",
        legend_font_style="normal",
        legend_font_size=10,
        legend_font_weight="normal",
        legend_marker_scale=None,
        legend_location=2,
        legend_bbox_to_anchor=(1.05, 1.0),
        legend_border_axes_pad=None,
        legend_n_columns=1,
        legend_horizontal_spacing=None,
        legend_vertical_spacing=None,
        legend_border=True,
        legend_border_padding=None,
        legend_shadow=False,
        legend_rounded_corners=False,
        render_axes=True,
        axes_font_name="sans-serif",
        axes_font_size=10,
        axes_font_style="normal",
        axes_font_weight="normal",
        figure_size=(7, 7),
        render_grid=True,
        grid_line_style="--",
        grid_line_width=1,
    ):
        import matplotlib.pyplot as plt

        # Check the viewer options that can be different for each plotted curve
        n_curves = len(self.y_axis)
        render_lines = _check_render_flag(
            render_lines,
            n_curves,
            "Must pass a list of different "
            "render_lines flag for each curve or "
            "a single render_lines flag for all "
            "curves.",
        )
        render_markers = _check_render_flag(
            render_markers,
            n_curves,
            "Must pass a list of different "
            "render_markers flag for each "
            "curve or a single render_markers "
            "flag for all curves.",
        )
        line_colour = _check_colours_list(
            True,
            line_colour,
            n_curves,
            "Must pass a list of line colours with length n_curves or a single "
            "line colour for all curves.",
        )
        line_style = _check_colours_list(
            True,
            line_style,
            n_curves,
            "Must pass a list of line styles with length n_curves or a single "
            "line style for all curves.",
        )
        line_width = _check_colours_list(
            True,
            line_width,
            n_curves,
            "Must pass a list of line widths with length n_curves or a single "
            "line width for all curves.",
        )
        marker_style = _check_colours_list(
            True,
            marker_style,
            n_curves,
            "Must pass a list of marker styles with length n_curves or a "
            "single marker style for all curves.",
        )
        marker_size = _check_colours_list(
            True,
            marker_size,
            n_curves,
            "Must pass a list of marker sizes with length n_curves or a single "
            "marker size for all curves.",
        )
        marker_face_colour = _check_colours_list(
            True,
            marker_face_colour,
            n_curves,
            "Must pass a list of marker face colours with length n_curves or a "
            "single marker face colour for all curves.",
        )
        marker_edge_colour = _check_colours_list(
            True,
            marker_edge_colour,
            n_curves,
            "Must pass a list of marker edge colours with length n_curves or a "
            "single marker edge colour for all curves.",
        )
        marker_edge_width = _check_colours_list(
            True,
            marker_edge_width,
            n_curves,
            "Must pass a list of marker edge widths with length n_curves or a "
            "single marker edge width for all curves.",
        )

        # plot all curves
        ax = plt.gca()
        for i, y in enumerate(self.y_axis):
            linestyle = line_style[i]
            if not render_lines[i]:
                linestyle = "None"
            marker = marker_style[i]
            if not render_markers[i]:
                marker = "None"
            plt.plot(
                self.x_axis,
                y,
                color=line_colour[i],
                linestyle=linestyle,
                linewidth=line_width[i],
                marker=marker,
                markeredgecolor=marker_edge_colour[i],
                markerfacecolor=marker_face_colour[i],
                markeredgewidth=marker_edge_width[i],
                markersize=marker_size[i],
                label=self.legend_entries[i],
            )

        # set legend
        _set_legend(
            ax,
            legend_handles=None,
            render_legend=render_legend,
            legend_title=legend_title,
            legend_font_name=legend_font_name,
            legend_font_style=legend_font_style,
            legend_font_size=legend_font_size,
            legend_font_weight=legend_font_weight,
            legend_marker_scale=legend_marker_scale,
            legend_location=legend_location,
            legend_bbox_to_anchor=legend_bbox_to_anchor,
            legend_border_axes_pad=legend_border_axes_pad,
            legend_n_columns=legend_n_columns,
            legend_horizontal_spacing=legend_horizontal_spacing,
            legend_vertical_spacing=legend_vertical_spacing,
            legend_border=legend_border,
            legend_border_padding=legend_border_padding,
            legend_shadow=legend_shadow,
            legend_rounded_corners=legend_rounded_corners,
        )

        # set axes options
        _set_axes_options(
            ax,
            render_axes=render_axes,
            inverted_y_axis=False,
            axes_font_name=axes_font_name,
            axes_font_size=axes_font_size,
            axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight,
            axes_x_limits=self.x_axis_limits,
            axes_y_limits=self.y_axis_limits,
            axes_x_ticks=self.x_axis_ticks,
            axes_y_ticks=self.y_axis_ticks,
            axes_x_label=self.x_label,
            axes_y_label=self.y_label,
            title=self.title,
        )

        # set grid options
        _set_grid_options(
            render_grid=render_grid,
            grid_line_style=grid_line_style,
            grid_line_width=grid_line_width,
        )

        # set figure size
        _set_figure_size(self.figure, figure_size)

        return self


class MatplotlibMultiImageViewer2d(MatplotlibRenderer):
    def __init__(self, figure_id, new_figure, image_list):
        super(MatplotlibMultiImageViewer2d, self).__init__(figure_id, new_figure)
        self.image_list = image_list

    def render(self, interval=50, **kwargs):
        import matplotlib.pyplot as plt
        import matplotlib.cm as cm
        import matplotlib.animation as animation

        if len(self.image_list[0].shape) == 2:
            # Single channels are viewed in Gray
            _ax = plt.imshow(self.image_list[0], cmap=cm.Greys_r, **kwargs)
        else:
            _ax = plt.imshow(self.image_list[0], **kwargs)

        def init():
            return (_ax,)

        def animate(j):
            _ax.set_data(self.image_list[j])
            return (_ax,)

        self._ani = animation.FuncAnimation(
            self.figure,
            animate,
            init_func=init,
            frames=len(self.image_list),
            interval=interval,
            blit=True,
        )
        return self


class MatplotlibMultiImageSubplotsViewer2d(MatplotlibRenderer, MatplotlibSubplots):
    def __init__(self, figure_id, new_figure, image_list):
        super(MatplotlibMultiImageSubplotsViewer2d, self).__init__(
            figure_id, new_figure
        )
        self.image_list = image_list
        self.num_subplots = self.image_list[0].shape[2]
        self.plot_layout = self._subplot_layout(self.num_subplots)

    def render(self, interval=50, **kwargs):
        import matplotlib.cm as cm
        import matplotlib.animation as animation
        import matplotlib.pyplot as plt

        p = self.plot_layout
        _axs = []
        for i in range(self.image_list[0].shape[2]):
            plt.subplot(p[0], p[1], 1 + i)
            # Hide the x and y labels
            plt.axis("off")
            _ax = plt.imshow(self.image_list[0][:, :, i], cmap=cm.Greys_r, **kwargs)
            _axs.append(_ax)

        def init():
            return _axs

        def animate(j):
            for k, _ax in enumerate(_axs):
                _ax.set_data(self.image_list[j][:, :, k])
            return _axs

        self._ani = animation.FuncAnimation(
            self.figure,
            animate,
            init_func=init,
            frames=len(self.image_list),
            interval=interval,
            blit=True,
        )
        return self


def sample_colours_from_colourmap(n_colours, colour_map):
    import matplotlib.pyplot as plt

    cm = plt.get_cmap(colour_map)
    return [cm(1.0 * i / n_colours)[:3] for i in range(n_colours)]


def _check_colours_list(render_flag, colours_list, n_objects, error_str):
    if render_flag:
        if colours_list is None:
            # sample colours from jet colour map
            colours_list = sample_colours_from_colourmap(n_objects, GLOBAL_CMAP)
        if isinstance(colours_list, list):
            if len(colours_list) == 1:
                colours_list *= n_objects
            elif len(colours_list) != n_objects:
                raise ValueError(error_str)
        else:
            colours_list = [colours_list] * n_objects
    else:
        colours_list = [None] * n_objects
    return colours_list


def _check_render_flag(render_flag, n_objects, error_str):
    if isinstance(render_flag, bool):
        render_flag = [render_flag] * n_objects
    elif isinstance(render_flag, list):
        if len(render_flag) == 1:
            render_flag *= n_objects
        elif len(render_flag) != n_objects:
            raise ValueError(error_str)
    else:
        raise ValueError(error_str)
    return render_flag



from collections import OrderedDict
import numpy as np

from ..base import validate_input, connectivity_from_array, labeller_func


@labeller_func(group_label="hand_ibug_39")
def hand_ibug_39_to_hand_ibug_39(pcloud):
    r"""
    Apply the IBUG 39-point semantic labels.

    The semantic labels applied are as follows:

      - thumb
      - index
      - middle
      - ring
      - pinky
      - palm
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 39
    validate_input(pcloud, n_expected_points)

    thumb_indices = np.arange(0, 5)
    index_indices = np.arange(5, 12)
    middle_indices = np.arange(12, 19)
    ring_indices = np.arange(19, 26)
    pinky_indices = np.arange(26, 33)
    palm_indices = np.hstack((np.array([32, 25, 18, 11, 33, 34, 4]), np.arange(35, 39)))

    thumb_connectivity = connectivity_from_array(thumb_indices, close_loop=False)
    index_connectivity = connectivity_from_array(index_indices, close_loop=False)
    middle_connectivity = connectivity_from_array(middle_indices, close_loop=False)
    ring_connectivity = connectivity_from_array(ring_indices, close_loop=False)
    pinky_connectivity = connectivity_from_array(pinky_indices, close_loop=False)
    palm_connectivity = connectivity_from_array(palm_indices, close_loop=True)

    all_connectivity = np.vstack(
        [
            thumb_connectivity,
            index_connectivity,
            middle_connectivity,
            ring_connectivity,
            pinky_connectivity,
            palm_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["thumb"] = thumb_indices
    mapping["index"] = index_indices
    mapping["middle"] = middle_indices
    mapping["ring"] = ring_indices
    mapping["pinky"] = pinky_indices
    mapping["palm"] = palm_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises
from menpo.shape import PointCloud

from menpo.transform import (
    Homogeneous,
    Scale,
    NonUniformScale,
    UniformScale,
    Translation,
    Similarity,
    Rotation,
    AlignmentUniformScale,
    AlignmentRotation,
)

# NON-INPLACE COMPOSE


# 1a. Homogenous before/after with subclasses. All should promote to Homogeneous


def test_homog_compose_before_nonuniformscale():
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    s = NonUniformScale([3, 4])
    res = homog.compose_before(s)
    assert type(res) == Homogeneous
    assert_allclose(res.h_matrix, np.array([[0, 3, 0], [4, 0, 0], [0, 0, 1]]))


def test_homog_compose_after_uniformscale():
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    s = UniformScale(3, 2)
    res = homog.compose_after(s)
    assert type(res) == Homogeneous
    assert_allclose(res.h_matrix, np.array([[0, 3, 0], [3, 0, 0], [0, 0, 1]]))


def test_rotation_compose_before_homog():
    # can't do this inplace - so should just give transform chain
    rotation = Rotation(np.array([[1, 0], [0, 1]]))
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    res = rotation.compose_before(homog)
    assert type(res) == Homogeneous


def test_translation_compose_after_homog():
    # can't do this inplace - so should just give transform chain
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    t = Translation([3, 4])
    res = t.compose_after(homog)
    assert type(res) == Homogeneous


# 1b. Homogeneous composed with Alignment subclasses. Should loose alignment
# traits.


def test_homog_compose_before_alignment_nonuniformscale():
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    scale = UniformScale(2.5, 2)
    source = PointCloud(np.array([[0, 1], [1, 1], [-1, -5], [3, -5]]))
    target = scale.apply(source)
    # estimate the transform from source and target
    s = AlignmentUniformScale(source, target)
    res = homog.compose_before(s)
    assert type(res) == Homogeneous


def test_homog_compose_after_alignment_rotation():
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    source = PointCloud(np.array([[0, 1], [1, 1], [-1, -5], [3, -5]]))
    r = AlignmentRotation(source, source)
    res = homog.compose_after(r)
    assert type(res) == Homogeneous


def test_scale_compose_after_inplace_homog():
    # can't do this inplace - so should just give transform chain
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    s = Scale([3, 4])
    with raises(ValueError):
        s.compose_after_inplace(homog)


def test_homog_compose_after_inplace_scale():
    # this should be fine
    homog = Homogeneous(np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]))
    s = Scale([3, 4])
    homog.compose_after_inplace(s)
    assert_allclose(homog.h_matrix, np.array([[0, 4, 0], [3, 0, 0], [0, 0, 1]]))


def test_uniformscale_compose_after_translation():
    t = Translation([3, 4])
    s = UniformScale(2, 2)
    res = s.compose_after(t)
    assert type(res) == Similarity


def test_translation_compose_after_uniformscale():
    t = Translation([3, 4])
    s = UniformScale(2, 2)
    res = t.compose_after(s)
    assert type(res) == Similarity



from .landmark import lm2_importer, ljson_importer
from .image import pillow_importer, abs_importer, flo_importer
from .video import ffmpeg_types, ffmpeg_importer
from .landmark_image import asf_image_importer, pts_image_importer
from .pickle import pickle_importer, pickle_gzip_importer


image_types = {
    ".bmp": pillow_importer,
    ".dib": pillow_importer,
    ".dcx": pillow_importer,
    ".eps": pillow_importer,
    ".ps": pillow_importer,
    ".gif": ffmpeg_importer,
    ".im": pillow_importer,
    ".jpg": pillow_importer,
    ".jpg2": pillow_importer,
    ".jpx": pillow_importer,
    ".jpe": pillow_importer,
    ".jpeg": pillow_importer,
    ".pcd": pillow_importer,
    ".pcx": pillow_importer,
    ".png": pillow_importer,
    ".pbm": pillow_importer,
    ".pgm": pillow_importer,
    ".ppm": pillow_importer,
    ".psd": pillow_importer,
    ".tif": pillow_importer,
    ".tiff": pillow_importer,
    ".xbm": pillow_importer,
    ".xpm": pillow_importer,
    ".abs": abs_importer,
    ".flo": flo_importer,
}


ffmpeg_video_types = ffmpeg_types()
ffmpeg_video_types[".gif"] = ffmpeg_importer

image_landmark_types = {
    ".asf": asf_image_importer,
    ".lm2": lm2_importer,
    ".pts": pts_image_importer,
    ".ptsx": pts_image_importer,
    ".ljson": ljson_importer,
}

pickle_types = {".pkl": pickle_importer, ".pkl.gz": pickle_gzip_importer}



from unittest.mock import PropertyMock

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal
from pytest import raises

import menpo
import menpo.io as mio
from menpo.image import BooleanImage, Image, MaskedImage, OutOfMaskSampleError
from menpo.image.interpolation import cv2_perspective_interpolation
from menpo.shape import PointCloud, bounding_box
from menpo.transform import Affine, Rotation, UniformScale

CROP_COORDS = (np.array([70, 30]), np.array([169, 129]))


@pytest.fixture()
def interpolation_method(mocker, method):
    # This uses the pytest-mock package in order to provide a fixture that will
    # mock the global variable in the image module that determines if OpenCV
    # should be used or not. Useful for running the unit tests with and without
    # opencv
    interp = PropertyMock()
    interp.return_value = method
    mocker.patch("menpo.image.base.cv2_perspective_interpolation", new_callable=interp)


def opencv_and_scipy_interpolation(func):
    # This decorator combines the "interpolation_method" fixture from above
    # with a parameterize decorator to run unit tests with and without OpenCV
    parameterize = pytest.mark.parametrize(
        "method",
        [None, cv2_perspective_interpolation],
        ids=lambda x: "opencv" if x is not None else "scipy",
    )
    use_fixtures = pytest.mark.usefixtures("interpolation_method")
    return use_fixtures(parameterize(func))


@pytest.fixture()
def rgb_image():
    return mio.import_builtin_asset("takeo.ppm")


@pytest.fixture()
def gray_image():
    return mio.import_builtin_asset("takeo.ppm").as_greyscale()


@pytest.fixture()
def gray_template(gray_image):
    gray_template = gray_image.crop(*CROP_COORDS)
    mask = BooleanImage.init_blank(gray_template.shape)
    return gray_template, mask


@pytest.fixture()
def target_transform():
    initial_params = np.array([0, 0, 0, 0, 70, 30])
    return Affine.init_identity(2).from_vector(initial_params)


def test_warp_gray(gray_image, gray_template, target_transform):
    gray_template, mask = gray_template
    warped_im = gray_image.warp_to_mask(mask, target_transform)

    assert warped_im.shape == gray_template.shape
    assert_allclose(warped_im.pixels, gray_template.pixels)


def test_warp_gray_batch(gray_image, gray_template, target_transform):
    gray_template, mask = gray_template
    warped_im = gray_image.warp_to_mask(mask, target_transform, batch_size=100)

    assert warped_im.shape == gray_template.shape
    assert_allclose(warped_im.pixels, gray_template.pixels)


def test_warp_multi(rgb_image, target_transform):
    rgb_template = rgb_image.crop(*CROP_COORDS)
    mask = BooleanImage.init_blank(rgb_template.shape)
    warped_im = rgb_image.warp_to_mask(mask, target_transform)

    assert warped_im.shape == rgb_template.shape
    assert_allclose(warped_im.pixels, rgb_template.pixels)


def test_warp_to_mask_boolean():
    b = BooleanImage.init_blank((10, 10))
    b.pixels[:, :5] = False
    template_mask = BooleanImage.init_blank((10, 10))
    template_mask.pixels[:5, :] = False
    t = Affine.init_identity(2)
    warped_mask = b.warp_to_mask(template_mask, t)
    assert type(warped_mask) == BooleanImage
    result = template_mask.pixels.copy()
    result[:, :5] = False
    assert np.all(result == warped_mask.pixels)


def test_warp_to_mask_image():
    img = Image.init_blank((10, 10), n_channels=2)
    img.pixels[:, :, :5] = 0.5
    template_mask = BooleanImage.init_blank((10, 10))
    template_mask.pixels[:, 5:, :] = False
    t = Affine.init_identity(2)
    warped_img = img.warp_to_mask(template_mask, t)
    assert type(warped_img) == MaskedImage
    result = Image.init_blank((10, 10), n_channels=2).pixels
    result[:, :5, :5] = 0.5
    assert np.all(result == warped_img.pixels)


def test_warp_to_mask_masked_image():
    mask = BooleanImage.init_blank((15, 15))
    # make a truncated mask on the original image
    mask.pixels[0, -1, -1] = False
    img = MaskedImage.init_blank((15, 15), n_channels=2, mask=mask, fill=2.5)
    template_mask = BooleanImage.init_blank((10, 10), fill=False)
    template_mask.pixels[:, :5, :5] = True
    t = Affine.init_identity(2)
    warped_img = img.warp_to_mask(template_mask, t)
    assert type(warped_img) == MaskedImage

    result = Image.init_blank((10, 10), n_channels=2).pixels
    result[:, :5, :5] = 2.5
    result_mask = BooleanImage.init_blank((10, 10), fill=False).pixels
    result_mask[:, :5, :5] = True
    assert warped_img.n_true_pixels() == 25
    assert_allclose(result, warped_img.pixels)
    assert_allclose(result_mask, warped_img.mask.pixels)


def test_warp_to_mask_masked_image_all_true():
    img = MaskedImage.init_blank((10, 10), fill=2.5)

    template_mask = BooleanImage.init_blank((10, 10), fill=False)
    template_mask.pixels[:, :5, :5] = True
    t = Affine.init_identity(2)
    warped_img = img.warp_to_mask(template_mask, t)
    assert type(warped_img) == MaskedImage


def test_warp_to_shape_equal_warp_to_mask():
    r = menpo.transform.UniformScale(2.0, n_dims=2)
    b = mio.import_builtin_asset("breakingbad.jpg")
    m_shape = b.warp_to_shape((540, 960), r)
    m_mask = b.warp_to_mask(menpo.image.BooleanImage.init_blank((540, 960)), r)
    assert_allclose(m_shape.pixels, m_mask.pixels)


def test_warp_to_shape_batch():
    r = menpo.transform.Affine.init_identity(2)
    b = mio.import_builtin_asset("takeo.ppm")
    m_shape = b.warp_to_shape(b.shape, r, batch_size=100)
    assert_allclose(m_shape.pixels, b.pixels)


def test_rescale_boolean():
    mask = BooleanImage.init_blank((100, 100))
    mask.resize((10, 10))


def test_rescale_return_transform():
    img = Image.init_blank((100, 100), n_channels=1)
    img.landmarks["test"] = bounding_box([40, 40], [80, 80])
    cropped_img, transform = img.rescale(1.5, return_transform=True)
    img_back = cropped_img.warp_to_shape(img.shape, transform.pseudoinverse())
    assert_allclose(img_back.shape, img.shape)
    assert_allclose(img_back.pixels, img.pixels)
    assert_allclose(img_back.landmarks["test"].points, img.landmarks["test"].points)


def test_sample_image():
    im = Image.init_blank((100, 100), fill=2)
    p = PointCloud(np.array([[0, 0], [1, 0]]))

    arr = im.sample(p)
    assert_allclose(arr, [[2.0, 2.0]])


def test_sample_maskedimage():
    im = MaskedImage.init_blank((100, 100), fill=2)
    p = PointCloud(np.array([[0, 0], [1, 0]]))

    arr = im.sample(p)
    assert_allclose(arr, [[2.0, 2.0]])


def test_sample_maskedimage_error():
    m = np.zeros([100, 100], dtype=np.bool)
    im = MaskedImage.init_blank((100, 100), mask=m, fill=2)
    p = PointCloud(np.array([[0, 0], [1, 0]]))
    with raises(OutOfMaskSampleError):
        im.sample(p, verify_mask=True)


def test_sample_maskedimage_error_values():
    m = np.zeros([100, 100], dtype=np.bool)
    m[1, 0] = True
    im = MaskedImage.init_blank((100, 100), mask=m, fill=2)
    p = PointCloud(np.array([[0, 0], [1, 0]]))
    try:
        im.sample(p, verify_mask=True)
        # Expect exception!
        assert 0
    except OutOfMaskSampleError as e:
        sampled_mask = e.sampled_mask
        sampled_values = e.sampled_values
        assert_allclose(sampled_values, [[2.0, 2.0]])
        assert_allclose(sampled_mask, [[False, True]])


def test_sample_booleanimage():
    im = BooleanImage.init_blank((100, 100))
    im.pixels[0, 1, 0] = False
    p = PointCloud(np.array([[0, 0], [1, 0]]))

    arr = im.sample(p)
    assert_allclose(arr, [[True, False]])


@opencv_and_scipy_interpolation
def test_transform_about_centre(method):
    pixels_16 = np.arange(16, dtype=np.float)
    image = Image(pixels_16.reshape(4, 4))
    transform = Rotation.init_from_2d_ccw_angle(180).compose_before(
        UniformScale(2, n_dims=2)
    )
    # rotate 180 + scale degrees
    transformed_img = image.transform_about_centre(transform, mode="nearest", order=1)
    expected_pixels = np.concatenate(
        [np.linspace(15 - 2 * i, 12 - 2 * i, num=7)[None] for i in range(7)]
    )

    assert transformed_img.shape == (7, 7)
    assert_allclose(transformed_img.pixels[0], expected_pixels, rtol=1e-8, atol=1e-8)


@opencv_and_scipy_interpolation
def test_zoom_image():
    im = Image.init_blank((100, 100), fill=0)
    # White square in the centre of size 10x10
    im.pixels[0, 45:55, 45:55] = 1.0

    # Zoom in 50% makes the white square 5 pixel bigger in theory (16x16)
    zim = im.zoom(1.5)
    assert np.count_nonzero(zim.pixels) == 256


def test_zoom_booleanimage():
    im = BooleanImage.init_blank((100, 100))
    im.pixels[0, 0, :] = False
    im.pixels[0, -1, :] = False
    im.pixels[0, :, 0] = False
    im.pixels[0, :, -1] = False

    zim = im.zoom(1.2)
    assert np.all(zim.pixels)


@opencv_and_scipy_interpolation
def test_mirror_horizontal_image():
    image = Image(
        np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0], [9.0, 10.0, 11.0, 12.0]])
    )
    image.landmarks["temp"] = PointCloud(
        np.array([[1.0, 1.0], [1.0, 2.0], [2.0, 1.0], [2.0, 2.0]])
    )
    mirrored_img = image.mirror(axis=0)
    assert_allclose(
        mirrored_img.pixels,
        np.array(
            [[[9.0, 10.0, 11.0, 12.0], [5.0, 6.0, 7.0, 8.0], [1.0, 2.0, 3.0, 4.0]]]
        ),
    )
    assert_allclose(
        mirrored_img.landmarks["temp"].points,
        np.array([[1.0, 1.0], [1.0, 2.0], [0.0, 1.0], [0.0, 2.0]]),
    )


@opencv_and_scipy_interpolation
def test_mirror_vertical_image():
    image = Image(
        np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0], [9.0, 10.0, 11.0, 12.0]])
    )
    image.landmarks["temp"] = PointCloud(
        np.array([[1.0, 0.0], [1.0, 1.0], [2.0, 1.0], [2.0, 2.0]])
    )
    mirrored_img = image.mirror()
    assert_allclose(
        mirrored_img.pixels,
        np.array(
            [[[4.0, 3.0, 2.0, 1.0], [8.0, 7.0, 6.0, 5.0], [12.0, 11.0, 10.0, 9.0]]]
        ),
    )
    assert_allclose(
        mirrored_img.landmarks["temp"].points,
        np.array([[1.0, 3.0], [1.0, 2.0], [2.0, 2.0], [2.0, 1.0]]),
    )


def test_mirror_image_axis_error():
    with raises(ValueError):
        Image(np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])).mirror(axis=2)


def test_mirror_masked_image():
    image = MaskedImage(np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]]))
    mirrored_img = image.mirror()
    assert type(mirrored_img) == MaskedImage


def test_mirror_return_transform():
    img = Image.init_blank((100, 100), n_channels=1)
    img.landmarks["test"] = bounding_box([40, 40], [80, 80])
    cropped_img, transform = img.mirror(return_transform=True)
    img_back = cropped_img.warp_to_shape(img.shape, transform.pseudoinverse())
    assert_allclose(img_back.shape, img.shape)
    assert_allclose(img_back.pixels, img.pixels)
    assert_allclose(img_back.landmarks["test"].points, img.landmarks["test"].points)


@opencv_and_scipy_interpolation
def test_rotate_image_90_180():
    image = Image(
        np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0], [9.0, 10.0, 11.0, 12.0]])
    )
    image.landmarks["temp"] = PointCloud(
        np.array([[1.0, 1.0], [1.0, 2.0], [2.0, 1.0], [2.0, 2.0]])
    )
    # rotate 90 degrees
    rotated_img = image.rotate_ccw_about_centre(theta=90, order=1, mode="nearest")
    rotated_img.landmarks["temp"] = rotated_img.landmarks["temp"].constrain_to_bounds(
        rotated_img.bounds()
    )
    assert_allclose(
        rotated_img.pixels,
        np.array(
            [[[4.0, 8.0, 12.0], [3.0, 7.0, 11.0], [2.0, 6.0, 10.0], [1.0, 5.0, 9.0]]]
        ),
    )
    assert_almost_equal(
        rotated_img.landmarks["temp"].points,
        np.array([[2.0, 1.0], [1.0, 1.0], [2.0, 2.0], [1.0, 2.0]]),
    )

    # rotate 180 degrees
    rotated_img = image.rotate_ccw_about_centre(theta=180, order=1, mode="nearest")
    rotated_img.landmarks["temp"] = rotated_img.landmarks["temp"].constrain_to_bounds(
        rotated_img.bounds()
    )
    assert_allclose(
        rotated_img.pixels,
        np.array(
            [[[12.0, 11.0, 10.0, 9.0], [8.0, 7.0, 6.0, 5.0], [4.0, 3.0, 2.0, 1.0]]]
        ),
    )
    assert_almost_equal(
        rotated_img.landmarks["temp"].points,
        np.array([[1.0, 2.0], [1.0, 1.0], [0.0, 2.0], [0.0, 1.0]]),
    )


def test_rotate_image_45():
    image = Image(
        np.array(
            [
                [1.0, 2.0, 3.0, 4.0],
                [5.0, 6.0, 7.0, 8.0],
                [9.0, 10.0, 11.0, 12.0],
                [13.0, 14.0, 15.0, 16.0],
            ]
        )
    )
    image.landmarks["temp"] = PointCloud(
        np.array([[1.0, 1.0], [1.0, 2.0], [2.0, 1.0], [2.0, 2.0]])
    )
    rotated_img = image.rotate_ccw_about_centre(theta=45, order=0)
    assert_allclose(
        rotated_img.pixels,
        np.array(
            [
                [
                    [0.0, 0.0, 4.0, 0.0, 0.0],
                    [0.0, 3.0, 7.0, 8.0, 0.0],
                    [1.0, 6.0, 7.0, 11.0, 16.0],
                    [0.0, 5.0, 10.0, 15.0, 15.0],
                    [0.0, 0.0, 13.0, 14.0, 0.0],
                ]
            ]
        ),
    )
    assert_almost_equal(
        rotated_img.landmarks["temp"].points,
        np.array([[2.121, 1.414], [1.414, 2.121], [2.828, 2.121], [2.121, 2.828]]),
        decimal=3,
    )


def test_rotate_return_transform():
    img = Image.init_blank((100, 100), n_channels=1)
    img.landmarks["test"] = bounding_box([40, 40], [80, 80])
    cropped_img, transform = img.rotate_ccw_about_centre(60, return_transform=True)
    img_back = cropped_img.warp_to_shape(img.shape, transform.pseudoinverse())
    assert_allclose(img_back.shape, img.shape)
    assert_allclose(img_back.pixels, img.pixels)
    assert_allclose(img_back.landmarks["test"].points, img.landmarks["test"].points)


@opencv_and_scipy_interpolation
def test_maskedimage_retain_shape():
    image = mio.import_builtin_asset("takeo.ppm")
    image = image.as_masked()
    rotated_img = image.rotate_ccw_about_centre(theta=77, retain_shape=True)
    assert image.shape == rotated_img.shape
    assert type(rotated_img) == MaskedImage



from .input import (
    import_image,
    import_images,
    image_paths,
    import_video,
    import_videos,
    video_paths,
    import_landmark_file,
    import_landmark_files,
    landmark_file_paths,
    import_pickle,
    import_pickles,
    pickle_paths,
    import_builtin_asset,
    data_dir_path,
    data_path_to,
    ls_builtin_assets,
    register_image_importer,
    register_landmark_importer,
    register_pickle_importer,
    register_video_importer,
)
from .output import export_image, export_video, export_landmark_file, export_pickle
from .exceptions import OverwriteError



from .pointcloud import PointCloud, bounding_box, bounding_cuboid
from .mesh import TriMesh, ColouredTriMesh, TexturedTriMesh
from .groupops import mean_pointcloud
from .graph import (
    UndirectedGraph,
    DirectedGraph,
    Tree,
    PointUndirectedGraph,
    PointDirectedGraph,
    PointTree,
)
from .graph_predefined import (
    empty_graph,
    star_graph,
    complete_graph,
    chain_graph,
    delaunay_graph,
    stencil_grid,
)
from .labelled import LabelledPointUndirectedGraph



from collections import OrderedDict
import numpy as np
from scipy.sparse import csr_matrix
from pytest import raises
from numpy.testing import assert_allclose, assert_equal

from menpo.landmark import LandmarkManager
from menpo.shape import PointCloud, LabelledPointUndirectedGraph
from menpo.testing import is_same_array


points = np.ones((10, 3))
adjacency_matrix = csr_matrix((10, 10))
mask_dict = OrderedDict([("all", np.ones(10, dtype=np.bool))])


def test_LandmarkManager_set_LabelledPointUndirectedGraph():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)

    man = LandmarkManager()
    man["test_set"] = lgroup
    assert not is_same_array(man["test_set"].points, lgroup.points)
    assert man["test_set"]._labels_to_masks is not lgroup._labels_to_masks


def test_LandmarkManager_set_None_key():
    lgroup = LabelledPointUndirectedGraph.init_with_all_label(points, adjacency_matrix)

    man = LandmarkManager()
    with raises(ValueError):
        man[None] = lgroup


def test_LandmarkManager_set_pointcloud():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud

    new_pcloud = man["test_set"]
    assert new_pcloud is not pcloud
    assert isinstance(new_pcloud, PointCloud)


def test_LandmarkManager_copy_method():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud
    man_copy = man.copy()

    assert man_copy["test_set"] is not man["test_set"]
    assert not is_same_array(man_copy["test_set"].points, man["test_set"].points)


def test_LandmarkManager_set_PointCloud_not_copy_target():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud
    assert not is_same_array(man["test_set"].points, pcloud.points)
    assert_allclose(man["test_set"].points, np.ones([10, 3]))


def test_LandmarkManager_iterate():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud
    man["test_set2"] = pcloud
    man["test_set3"] = pcloud

    for l in man:
        assert not is_same_array(man[l].points, pcloud.points)
    assert len(man) == 3


def test_LandmarkManager_get():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)

    man = LandmarkManager()
    man._landmark_groups["test_set"] = lgroup

    assert man["test_set"] is lgroup
    assert is_same_array(man["test_set"].points, lgroup.points)


def test_LandmarkManager_set():
    lgroup = LabelledPointUndirectedGraph(points, adjacency_matrix, mask_dict)

    man = LandmarkManager()
    man["test_set"] = lgroup

    assert_allclose(man._landmark_groups["test_set"].points, lgroup.points)
    assert man._landmark_groups["test_set"].n_labels == 1


def test_LandmarkManager_del():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud

    del man["test_set"]

    assert man.n_groups == 0


def test_LandmarkManager_group_order():
    points = np.ones((10, 3))
    pcloud = PointCloud(points, copy=False)
    target = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud.copy()
    man["abc"] = pcloud.copy()
    man["def"] = pcloud.copy()
    assert_equal(list(man._landmark_groups.keys()), ["test_set", "abc", "def"])
    # check that after deleting and inserting the order will remain the same.
    del man["test_set"]
    man["tt"] = pcloud.copy()
    assert_equal(list(man._landmark_groups.keys()), ["abc", "def", "tt"])


def test_LandmarkManager_in():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud

    assert "test_set" in man


def test_LandmarkManager_str():
    pcloud = PointCloud(points)

    man = LandmarkManager()
    man["test_set"] = pcloud

    out_str = man.__str__()
    assert len(out_str) > 0



import numpy as np
from menpo.shape import TriMesh


def utils_mesh():
    trilist = np.array([[0, 1, 2], [0, 2, 3]])
    points = np.array([[0, 0], [1, 0], [1, 1], [0, 1]])
    return TriMesh(points, trilist=trilist)


gt_edge_indices = np.array([[0, 1], [1, 2], [2, 0], [0, 2], [2, 3], [3, 0]])

gt_edge_vectors = np.array([[1, 0], [0, 1], [1, 1], [1, 1], [-1, 0], [0, 1]])

gt_unique_edge_vectors = np.array([[1, 0], [1, 1], [0, 1], [0, 1], [-1, 0]])

gt_edge_lengths = np.array([1.0, 1.0, 1.41421356, 1.41421356, 1.0, 1.0])

gt_unique_edge_lengths = np.array([1.0, 1.41421356, 1.0, 1.0, 1.0])

gt_tri_areas = np.array([0.5, 0.5])


def test_edge_indices():
    np.testing.assert_allclose(utils_mesh().edge_indices(), gt_edge_indices)


def test_edge_vectors():
    np.testing.assert_allclose(utils_mesh().edge_vectors(), gt_edge_vectors)


def test_unique_edge_vectors():
    np.testing.assert_allclose(
        utils_mesh().unique_edge_vectors(), gt_unique_edge_vectors
    )


def test_edge_lengths():
    assert np.allclose(utils_mesh().edge_lengths(), gt_edge_lengths)


def test_unique_edge_lengths():
    np.testing.assert_allclose(
        utils_mesh().unique_edge_lengths(), gt_unique_edge_lengths
    )


def test_tri_areas():
    np.testing.assert_allclose(utils_mesh().tri_areas(), gt_tri_areas)


def test_2d_trimesh_2d_positive_areas():
    t = TriMesh(
        np.array([[0, 0], [0, 1], [1, 1], [1, 0]], dtype=np.float),
        trilist=np.array([[0, 2, 3], [0, 2, 1]]),
    )
    assert np.all(t.tri_areas() > 0)


def test_mean_tri_area():
    assert np.isclose(utils_mesh().mean_tri_area(), 0.5)


def test_mean_edge_length():
    np.testing.assert_allclose(
        utils_mesh().mean_edge_length(), np.mean(gt_unique_edge_lengths)
    )


def test_mean_edge_length_not_unique():
    np.testing.assert_allclose(
        utils_mesh().mean_edge_length(unique=False), np.mean(gt_edge_lengths)
    )



from collections import OrderedDict
import numpy as np

from ..base import (
    validate_input,
    connectivity_from_array,
    pcloud_and_lgroup_from_ranges,
    connectivity_from_range,
    labeller_func,
)


@labeller_func(group_label="face_ibug_68")
def face_ibug_68_to_face_ibug_68(pcloud):
    r"""
    Apply the IBUG 68-point semantic labels.

    The semantic labels are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 68
    validate_input(pcloud, n_expected_points)

    jaw_indices = np.arange(0, 17)
    lbrow_indices = np.arange(17, 22)
    rbrow_indices = np.arange(22, 27)
    upper_nose_indices = np.arange(27, 31)
    lower_nose_indices = np.arange(31, 36)
    leye_indices = np.arange(36, 42)
    reye_indices = np.arange(42, 48)
    outer_mouth_indices = np.arange(48, 60)
    inner_mouth_indices = np.arange(60, 68)

    jaw_connectivity = connectivity_from_array(jaw_indices)
    lbrow_connectivity = connectivity_from_array(lbrow_indices)
    rbrow_connectivity = connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack(
        [
            connectivity_from_array(upper_nose_indices),
            connectivity_from_array(lower_nose_indices),
        ]
    )
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack(
        [
            connectivity_from_array(outer_mouth_indices, close_loop=True),
            connectivity_from_array(inner_mouth_indices, close_loop=True),
        ]
    )

    all_connectivity = np.vstack(
        [
            jaw_connectivity,
            lbrow_connectivity,
            rbrow_connectivity,
            nose_connectivity,
            leye_connectivity,
            reye_connectivity,
            mouth_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["jaw"] = jaw_indices
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["nose"] = np.hstack((upper_nose_indices, lower_nose_indices))
    mapping["left_eye"] = leye_indices
    mapping["right_eye"] = reye_indices
    mapping["mouth"] = np.hstack((outer_mouth_indices, inner_mouth_indices))

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_68")
def face_ibug_68_mirrored_to_face_ibug_68(pcloud):
    r"""
    Apply the IBUG 68-point semantic labels, on a pointcloud that has been
    mirrored around the vertical axis (flipped around the Y-axis). Thus, on
    the flipped image the jaw etc would be the wrong way around. This
    rectifies that and returns a new PointCloud whereby all the points
    are oriented correctly.

    The semantic labels applied are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    new_pcloud, old_map = face_ibug_68_to_face_ibug_68(pcloud, return_mapping=True)
    lms_map = np.hstack(
        [
            old_map["jaw"][::-1],
            old_map["right_eyebrow"][::-1],
            old_map["left_eyebrow"][::-1],
            old_map["nose"][:4],
            old_map["nose"][4:][::-1],
            np.roll(old_map["right_eye"][::-1], 4),
            np.roll(old_map["left_eye"][::-1], 4),
            np.roll(old_map["mouth"][:12][::-1], 7),
            np.roll(old_map["mouth"][12:][::-1], 5),
        ]
    )
    return new_pcloud.from_vector(pcloud.points[lms_map]), old_map


@labeller_func(group_label="face_ibug_66")
def face_ibug_68_to_face_ibug_66(pcloud):
    r"""
    Apply the IBUG 66-point semantic labels, but ignoring the 2 points
    describing the inner mouth corners).

    The semantic labels applied are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 68
    validate_input(pcloud, n_expected_points)

    jaw_indices = np.arange(0, 17)
    lbrow_indices = np.arange(17, 22)
    rbrow_indices = np.arange(22, 27)
    upper_nose_indices = np.arange(27, 31)
    lower_nose_indices = np.arange(31, 36)
    leye_indices = np.arange(36, 42)
    reye_indices = np.arange(42, 48)
    outer_mouth_indices = np.arange(48, 60)
    inner_mouth_indices = np.hstack((48, np.arange(60, 63), 54, np.arange(63, 66)))

    jaw_connectivity = connectivity_from_array(jaw_indices)
    lbrow_connectivity = connectivity_from_array(lbrow_indices)
    rbrow_connectivity = connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack(
        [
            connectivity_from_array(upper_nose_indices),
            connectivity_from_array(lower_nose_indices),
        ]
    )
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack(
        [
            connectivity_from_array(outer_mouth_indices, close_loop=True),
            connectivity_from_array(inner_mouth_indices, close_loop=True),
        ]
    )

    all_connectivity = np.vstack(
        [
            jaw_connectivity,
            lbrow_connectivity,
            rbrow_connectivity,
            nose_connectivity,
            leye_connectivity,
            reye_connectivity,
            mouth_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["jaw"] = jaw_indices
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["nose"] = np.hstack([upper_nose_indices, lower_nose_indices])
    mapping["left_eye"] = leye_indices
    mapping["right_eye"] = reye_indices
    mapping["mouth"] = np.hstack([outer_mouth_indices, inner_mouth_indices])

    # Ignore the two inner mouth points
    ind = np.hstack((np.arange(60), np.arange(61, 64), np.arange(65, 68)))
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_51")
def face_ibug_68_to_face_ibug_51(pcloud):
    r"""
    Apply the IBUG 51-point semantic labels, but removing the annotations
    corresponding to the jaw region.

    The semantic labels applied are as follows:

      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 68
    validate_input(pcloud, n_expected_points)

    lbrow_indices = np.arange(0, 5)
    rbrow_indices = np.arange(5, 10)
    upper_nose_indices = np.arange(10, 14)
    lower_nose_indices = np.arange(14, 19)
    leye_indices = np.arange(19, 25)
    reye_indices = np.arange(25, 31)
    outer_mouth_indices = np.arange(31, 43)
    inner_mouth_indices = np.arange(43, 51)

    lbrow_connectivity = connectivity_from_array(lbrow_indices)
    rbrow_connectivity = connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack(
        [
            connectivity_from_array(upper_nose_indices),
            connectivity_from_array(lower_nose_indices),
        ]
    )
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack(
        [
            connectivity_from_array(outer_mouth_indices, close_loop=True),
            connectivity_from_array(inner_mouth_indices, close_loop=True),
        ]
    )

    all_connectivity = np.vstack(
        [
            lbrow_connectivity,
            rbrow_connectivity,
            nose_connectivity,
            leye_connectivity,
            reye_connectivity,
            mouth_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["nose"] = np.hstack([upper_nose_indices, lower_nose_indices])
    mapping["left_eye"] = leye_indices
    mapping["right_eye"] = reye_indices
    mapping["mouth"] = np.hstack([outer_mouth_indices, inner_mouth_indices])

    # Ignore the two inner mouth points
    ind = np.arange(17, 68)
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_49")
def face_ibug_49_to_face_ibug_49(pcloud):
    r"""
    Apply the IBUG 49-point semantic labels.

    The semantic labels applied are as follows:

      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 49
    validate_input(pcloud, n_expected_points)

    lbrow_indices = np.arange(0, 5)
    rbrow_indices = np.arange(5, 10)
    upper_nose_indices = np.arange(10, 14)
    lower_nose_indices = np.arange(14, 19)
    leye_indices = np.arange(19, 25)
    reye_indices = np.arange(25, 31)
    outer_mouth_indices = np.arange(31, 43)
    inner_mouth_indices = np.hstack((31, np.arange(43, 46), 37, np.arange(46, 49)))

    lbrow_connectivity = connectivity_from_array(lbrow_indices)
    rbrow_connectivity = connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack(
        [
            connectivity_from_array(upper_nose_indices),
            connectivity_from_array(lower_nose_indices),
        ]
    )
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack(
        [
            connectivity_from_array(outer_mouth_indices, close_loop=True),
            connectivity_from_array(inner_mouth_indices, close_loop=True),
        ]
    )

    all_connectivity = np.vstack(
        [
            lbrow_connectivity,
            rbrow_connectivity,
            nose_connectivity,
            leye_connectivity,
            reye_connectivity,
            mouth_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["nose"] = np.hstack([upper_nose_indices, lower_nose_indices])
    mapping["left_eye"] = leye_indices
    mapping["right_eye"] = reye_indices
    mapping["mouth"] = np.hstack([outer_mouth_indices, inner_mouth_indices])

    # Ignore the two inner mouth points
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_49")
def face_ibug_68_to_face_ibug_49(pcloud):
    r"""
    Apply the IBUG 49-point semantic labels, but removing the annotations
    corresponding to the jaw region and the 2 describing the inner mouth
    corners.

    The semantic labels applied are as follows:

      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 68
    validate_input(pcloud, n_expected_points)

    lbrow_indices = np.arange(0, 5)
    rbrow_indices = np.arange(5, 10)
    upper_nose_indices = np.arange(10, 14)
    lower_nose_indices = np.arange(14, 19)
    leye_indices = np.arange(19, 25)
    reye_indices = np.arange(25, 31)
    outer_mouth_indices = np.arange(31, 43)
    inner_mouth_indices = np.hstack((31, np.arange(43, 46), 37, np.arange(46, 49)))

    lbrow_connectivity = connectivity_from_array(lbrow_indices)
    rbrow_connectivity = connectivity_from_array(rbrow_indices)
    nose_connectivity = np.vstack(
        [
            connectivity_from_array(upper_nose_indices),
            connectivity_from_array(lower_nose_indices),
        ]
    )
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    mouth_connectivity = np.vstack(
        [
            connectivity_from_array(outer_mouth_indices, close_loop=True),
            connectivity_from_array(inner_mouth_indices, close_loop=True),
        ]
    )

    all_connectivity = np.vstack(
        [
            lbrow_connectivity,
            rbrow_connectivity,
            nose_connectivity,
            leye_connectivity,
            reye_connectivity,
            mouth_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["nose"] = np.hstack([upper_nose_indices, lower_nose_indices])
    mapping["left_eye"] = leye_indices
    mapping["right_eye"] = reye_indices
    mapping["mouth"] = np.hstack([outer_mouth_indices, inner_mouth_indices])

    # Ignore the two inner mouth points
    ind = np.hstack((np.arange(17, 60), np.arange(61, 64), np.arange(65, 68)))
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_68_trimesh")
def face_ibug_68_to_face_ibug_68_trimesh(pcloud):
    r"""
    Apply the IBUG 68-point semantic labels, with trimesh connectivity.

    The semantic labels applied are as follows:

      - tri

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import TriMesh

    n_expected_points = 68
    validate_input(pcloud, n_expected_points)

    tri_list = np.array(
        [
            [47, 29, 28],
            [44, 43, 23],
            [38, 20, 21],
            [47, 28, 42],
            [49, 61, 60],
            [40, 41, 37],
            [37, 19, 20],
            [28, 40, 39],
            [38, 21, 39],
            [36, 1, 0],
            [48, 59, 4],
            [49, 60, 48],
            [67, 59, 60],
            [13, 53, 14],
            [61, 51, 62],
            [57, 8, 7],
            [52, 51, 33],
            [61, 67, 60],
            [52, 63, 51],
            [66, 56, 57],
            [35, 30, 29],
            [53, 52, 35],
            [37, 36, 17],
            [18, 37, 17],
            [37, 38, 40],
            [38, 37, 20],
            [19, 37, 18],
            [38, 39, 40],
            [28, 29, 40],
            [41, 36, 37],
            [27, 39, 21],
            [41, 31, 1],
            [30, 32, 31],
            [33, 51, 50],
            [33, 30, 34],
            [31, 40, 29],
            [36, 0, 17],
            [31, 2, 1],
            [31, 41, 40],
            [1, 36, 41],
            [31, 49, 2],
            [2, 49, 3],
            [60, 59, 48],
            [3, 49, 48],
            [31, 32, 50],
            [48, 4, 3],
            [59, 5, 4],
            [58, 67, 66],
            [5, 59, 58],
            [58, 59, 67],
            [7, 6, 58],
            [66, 57, 58],
            [13, 54, 53],
            [7, 58, 57],
            [6, 5, 58],
            [50, 61, 49],
            [62, 67, 61],
            [31, 50, 49],
            [32, 33, 50],
            [30, 33, 32],
            [34, 52, 33],
            [35, 52, 34],
            [53, 63, 52],
            [62, 63, 65],
            [62, 51, 63],
            [66, 65, 56],
            [63, 53, 64],
            [62, 66, 67],
            [62, 65, 66],
            [57, 56, 9],
            [65, 63, 64],
            [8, 57, 9],
            [9, 56, 10],
            [10, 56, 11],
            [11, 56, 55],
            [11, 55, 12],
            [56, 65, 55],
            [55, 64, 54],
            [55, 65, 64],
            [55, 54, 12],
            [64, 53, 54],
            [12, 54, 13],
            [45, 46, 44],
            [35, 34, 30],
            [14, 53, 35],
            [15, 46, 45],
            [27, 28, 39],
            [27, 42, 28],
            [35, 29, 47],
            [30, 31, 29],
            [15, 35, 46],
            [15, 14, 35],
            [43, 22, 23],
            [27, 21, 22],
            [24, 44, 23],
            [44, 47, 43],
            [43, 47, 42],
            [46, 35, 47],
            [26, 45, 44],
            [46, 47, 44],
            [25, 44, 24],
            [25, 26, 44],
            [16, 15, 45],
            [16, 45, 26],
            [22, 42, 43],
            [50, 51, 61],
            [27, 22, 42],
        ]
    )
    new_pcloud = TriMesh(pcloud.points, trilist=tri_list)

    mapping = OrderedDict()
    mapping["tri"] = np.arange(new_pcloud.n_points)

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_66_trimesh")
def face_ibug_68_to_face_ibug_66_trimesh(pcloud):
    r"""
    Apply the IBUG 66-point semantic labels, with trimesh connectivity.

    The semantic labels applied are as follows:

      - tri

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import TriMesh

    # Apply face_ibug_68_to_face_ibug_66
    new_pcloud = face_ibug_68_to_face_ibug_66(pcloud)

    # This is in terms of the 66 points
    tri_list = np.array(
        [
            [47, 29, 28],
            [44, 43, 23],
            [38, 20, 21],
            [47, 28, 42],
            [40, 41, 37],
            [51, 62, 61],
            [37, 19, 20],
            [28, 40, 39],
            [38, 21, 39],
            [36, 1, 0],
            [48, 59, 4],
            [49, 60, 48],
            [13, 53, 14],
            [60, 51, 61],
            [51, 51, 62],
            [52, 51, 33],
            [49, 50, 60],
            [57, 7, 8],
            [64, 56, 57],
            [35, 30, 29],
            [52, 62, 53],
            [53, 52, 35],
            [37, 36, 17],
            [18, 37, 17],
            [37, 38, 40],
            [38, 37, 20],
            [19, 37, 18],
            [38, 39, 40],
            [28, 29, 40],
            [41, 36, 37],
            [27, 39, 21],
            [41, 31, 1],
            [30, 32, 31],
            [33, 51, 50],
            [33, 30, 34],
            [31, 40, 29],
            [36, 0, 17],
            [31, 2, 1],
            [31, 41, 40],
            [1, 36, 41],
            [31, 49, 2],
            [2, 49, 3],
            [3, 49, 48],
            [31, 32, 50],
            [62, 53, 54],
            [48, 4, 3],
            [59, 5, 4],
            [58, 65, 64],
            [5, 59, 58],
            [58, 59, 65],
            [7, 6, 58],
            [64, 57, 58],
            [13, 54, 53],
            [7, 58, 57],
            [6, 5, 58],
            [63, 55, 54],
            [65, 59, 48],
            [31, 50, 49],
            [32, 33, 50],
            [30, 33, 32],
            [34, 52, 33],
            [35, 52, 34],
            [48, 60, 65],
            [64, 63, 56],
            [60, 65, 61],
            [65, 64, 61],
            [57, 56, 9],
            [8, 57, 9],
            [64, 63, 61],
            [9, 56, 10],
            [10, 56, 11],
            [11, 56, 55],
            [11, 55, 12],
            [56, 63, 55],
            [51, 52, 62],
            [55, 54, 12],
            [63, 54, 62],
            [61, 62, 63],
            [12, 54, 13],
            [45, 46, 44],
            [35, 34, 30],
            [14, 53, 35],
            [15, 46, 45],
            [27, 28, 39],
            [27, 42, 28],
            [35, 29, 47],
            [30, 31, 29],
            [15, 35, 46],
            [15, 14, 35],
            [43, 22, 23],
            [27, 21, 22],
            [24, 44, 23],
            [44, 47, 43],
            [43, 47, 42],
            [46, 35, 47],
            [26, 45, 44],
            [46, 47, 44],
            [25, 44, 24],
            [25, 26, 44],
            [16, 15, 45],
            [16, 45, 26],
            [22, 42, 43],
            [50, 60, 51],
            [27, 22, 42],
        ]
    )

    new_pcloud = TriMesh(new_pcloud.points, trilist=tri_list, copy=False)

    mapping = OrderedDict()
    mapping["tri"] = np.arange(new_pcloud.n_points)

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_51_trimesh")
def face_ibug_68_to_face_ibug_51_trimesh(pcloud):
    r"""
    Apply the IBUG 51-point semantic labels, with trimesh connectivity..

    The semantic labels applied are as follows:

      - tri

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import TriMesh

    # Apply face_ibug_68_to_face_ibug_51
    new_pcloud = face_ibug_68_to_face_ibug_51(pcloud)

    # This is in terms of the 51 points
    tri_list = np.array(
        [
            [30, 12, 11],
            [27, 26, 6],
            [21, 3, 4],
            [30, 11, 25],
            [32, 44, 43],
            [23, 24, 20],
            [20, 2, 3],
            [11, 23, 22],
            [21, 4, 22],
            [32, 43, 31],
            [50, 42, 43],
            [44, 34, 45],
            [35, 34, 16],
            [44, 50, 43],
            [35, 46, 34],
            [49, 39, 40],
            [18, 13, 12],
            [36, 35, 18],
            [20, 19, 0],
            [1, 20, 0],
            [20, 21, 23],
            [21, 20, 3],
            [2, 20, 1],
            [21, 22, 23],
            [11, 12, 23],
            [24, 19, 20],
            [10, 22, 4],
            [13, 15, 14],
            [16, 34, 33],
            [16, 13, 17],
            [14, 23, 12],
            [14, 24, 23],
            [43, 42, 31],
            [14, 15, 33],
            [41, 50, 49],
            [41, 42, 50],
            [49, 40, 41],
            [33, 44, 32],
            [45, 50, 44],
            [14, 33, 32],
            [15, 16, 33],
            [13, 16, 15],
            [17, 35, 16],
            [18, 35, 17],
            [36, 46, 35],
            [45, 46, 48],
            [45, 34, 46],
            [49, 48, 39],
            [46, 36, 47],
            [45, 49, 50],
            [45, 48, 49],
            [48, 46, 47],
            [39, 48, 38],
            [38, 47, 37],
            [38, 48, 47],
            [47, 36, 37],
            [28, 29, 27],
            [18, 17, 13],
            [10, 11, 22],
            [10, 25, 11],
            [18, 12, 30],
            [13, 14, 12],
            [26, 5, 6],
            [10, 4, 5],
            [7, 27, 6],
            [27, 30, 26],
            [26, 30, 25],
            [29, 18, 30],
            [9, 28, 27],
            [29, 30, 27],
            [8, 27, 7],
            [8, 9, 27],
            [5, 25, 26],
            [33, 34, 44],
            [10, 5, 25],
        ]
    )

    new_pcloud = TriMesh(new_pcloud.points, trilist=tri_list, copy=False)

    mapping = OrderedDict()
    mapping["tri"] = np.arange(new_pcloud.n_points)

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_49_trimesh")
def face_ibug_68_to_face_ibug_49_trimesh(pcloud):
    r"""
    Apply the IBUG 49-point semantic labels, with trimesh connectivity.

    The semantic labels applied are as follows:

      - tri

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import TriMesh

    # Apply face_ibug_68_to_face_ibug_49
    new_pcloud = face_ibug_68_to_face_ibug_49(pcloud)

    # This is in terms of the 49 points
    tri_list = np.array(
        [
            [47, 29, 28],
            [44, 43, 23],
            [38, 20, 21],
            [47, 28, 42],
            [40, 41, 37],
            [51, 62, 61],
            [37, 19, 20],
            [28, 40, 39],
            [38, 21, 39],
            [36, 1, 0],
            [48, 59, 4],
            [49, 60, 48],
            [13, 53, 14],
            [60, 51, 61],
            [51, 51, 62],
            [52, 51, 33],
            [49, 50, 60],
            [57, 7, 8],
            [64, 56, 57],
            [35, 30, 29],
            [52, 62, 53],
            [53, 52, 35],
            [37, 36, 17],
            [18, 37, 17],
            [37, 38, 40],
            [38, 37, 20],
            [19, 37, 18],
            [38, 39, 40],
            [28, 29, 40],
            [41, 36, 37],
            [27, 39, 21],
            [41, 31, 1],
            [30, 32, 31],
            [33, 51, 50],
            [33, 30, 34],
            [31, 40, 29],
            [36, 0, 17],
            [31, 2, 1],
            [31, 41, 40],
            [1, 36, 41],
            [31, 49, 2],
            [2, 49, 3],
            [3, 49, 48],
            [31, 32, 50],
            [62, 53, 54],
            [48, 4, 3],
            [59, 5, 4],
            [58, 65, 64],
            [5, 59, 58],
            [58, 59, 65],
            [7, 6, 58],
            [64, 57, 58],
            [13, 54, 53],
            [7, 58, 57],
            [6, 5, 58],
            [63, 55, 54],
            [65, 59, 48],
            [31, 50, 49],
            [32, 33, 50],
            [30, 33, 32],
            [34, 52, 33],
            [35, 52, 34],
            [48, 60, 65],
            [64, 63, 56],
            [60, 65, 61],
            [65, 64, 61],
            [57, 56, 9],
            [8, 57, 9],
            [64, 63, 61],
            [9, 56, 10],
            [10, 56, 11],
            [11, 56, 55],
            [11, 55, 12],
            [56, 63, 55],
            [51, 52, 62],
            [55, 54, 12],
            [63, 54, 62],
            [61, 62, 63],
            [12, 54, 13],
            [45, 46, 44],
            [35, 34, 30],
            [14, 53, 35],
            [15, 46, 45],
            [27, 28, 39],
            [27, 42, 28],
            [35, 29, 47],
            [30, 31, 29],
            [15, 35, 46],
            [15, 14, 35],
            [43, 22, 23],
            [27, 21, 22],
            [24, 44, 23],
            [44, 47, 43],
            [43, 47, 42],
            [46, 35, 47],
            [26, 45, 44],
            [46, 47, 44],
            [25, 44, 24],
            [25, 26, 44],
            [16, 15, 45],
            [16, 45, 26],
            [22, 42, 43],
            [50, 60, 51],
            [27, 22, 42],
        ]
    )

    new_pcloud = TriMesh(new_pcloud.points, trilist=tri_list, copy=False)

    mapping = OrderedDict()
    mapping["tri"] = np.arange(new_pcloud.n_points)

    return new_pcloud, mapping


@labeller_func(group_label="face_ibug_65")
def face_ibug_68_to_face_ibug_65(pcloud):
    r"""
    Apply the IBUG 68 point semantic labels, but ignore the 3 points that are
    coincident for a closed mouth (bottom of the inner mouth).

    The semantic labels applied are as follows:

      - jaw
      - left_eyebrow
      - right_eyebrow
      - nose
      - left_eye
      - right_eye
      - mouth

    References
    ----------
    .. [1] http://www.multipie.org/
    .. [2] http://ibug.doc.ic.ac.uk/resources/300-W/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    # Apply face_ibug_68_to_face_ibug_68
    new_pcloud, mapping = face_ibug_68_to_face_ibug_68(pcloud, return_mapping=True)

    # The coincident points are considered the final 3 landmarks (bottom of
    # the inner mouth points). We remove all the edges for the inner mouth
    # which are the last 8.
    edges = new_pcloud.edges[:-8]
    # Re-add the inner mouth without the bottom 3 points
    edges = np.vstack([edges, connectivity_from_range((60, 65), close_loop=True)])

    # Luckily, OrderedDict maintains the original ordering despite updates
    outer_mouth_indices = np.arange(48, 60)
    inner_mouth_indices = np.arange(60, 65)
    mapping["mouth"] = np.hstack([outer_mouth_indices, inner_mouth_indices])

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        new_pcloud.points[:-3], edges, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="face_imm_58")
def face_imm_58_to_face_imm_58(pcloud):
    r"""
    Apply the 58-point semantic labels from the IMM dataset.

    The semantic labels applied are as follows:

      - jaw
      - left_eye
      - right_eye
      - left_eyebrow
      - right_eyebrow
      - mouth
      - nose

    References
    ----------
    .. [1] http://www2.imm.dtu.dk/~aam/
    """
    n_expected_points = 58
    validate_input(pcloud, n_expected_points)

    labels = OrderedDict(
        [
            ("jaw", (0, 13, False)),
            ("left_eye", (13, 21, True)),
            ("right_eye", (21, 29, True)),
            ("left _eyebrow", (29, 34, False)),
            ("right_eyebrow", (34, 39, False)),
            ("mouth", (39, 47, True)),
            ("nose", (47, 58, False)),
        ]
    )
    return pcloud_and_lgroup_from_ranges(pcloud, labels)


@labeller_func(group_label="face_lfpw_29")
def face_lfpw_29_to_face_lfpw_29(pcloud):
    r"""
    Apply the 29-point semantic labels from the original LFPW dataset.

    The semantic labels applied are as follows:

      - chin
      - left_eye
      - right_eye
      - left_eyebrow
      - right_eyebrow
      - mouth
      - nose

    References
    ----------
    .. [1] http://homes.cs.washington.edu/~neeraj/databases/lfpw/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 29
    validate_input(pcloud, n_expected_points)

    chin_indices = np.array([28])
    outer_leye_indices = np.array([8, 12, 10, 13])
    pupil_leye_indices = np.array([16])
    outer_reye_indices = np.array([11, 14, 9, 15])
    pupil_reye_indices = np.array([17])
    lbrow_indices = np.array([0, 4, 2, 5])
    rbrow_indices = np.array([3, 6, 1, 7])
    outer_mouth_indices = np.array([22, 24, 23, 27])
    inner_mouth_indices = np.array([22, 25, 23, 26])
    nose_indices = np.array([18, 20, 19, 21])

    chin_connectivity = connectivity_from_array(chin_indices, close_loop=True)
    leye_connectivity = connectivity_from_array(outer_leye_indices, close_loop=True)
    reye_connectivity = connectivity_from_array(outer_reye_indices, close_loop=True)
    lbrow_connectivity = connectivity_from_array(lbrow_indices, close_loop=True)
    rbrow_connectivity = connectivity_from_array(rbrow_indices, close_loop=True)
    mouth_connectivity = np.vstack(
        [
            connectivity_from_array(outer_mouth_indices, close_loop=True),
            connectivity_from_array(inner_mouth_indices, close_loop=True),
        ]
    )
    nose_connectivity = connectivity_from_array(nose_indices, close_loop=True)

    all_connectivity = np.vstack(
        [
            chin_connectivity,
            leye_connectivity,
            reye_connectivity,
            lbrow_connectivity,
            rbrow_connectivity,
            mouth_connectivity,
            nose_connectivity,
        ]
    )

    mapping = OrderedDict()
    mapping["chin"] = chin_indices
    mapping["left_eye"] = np.hstack((outer_leye_indices, pupil_leye_indices))
    mapping["right_eye"] = np.hstack((outer_reye_indices, pupil_reye_indices))
    mapping["left_eyebrow"] = lbrow_indices
    mapping["right_eyebrow"] = rbrow_indices
    mapping["mouth"] = np.hstack((outer_mouth_indices, inner_mouth_indices))
    mapping["nose"] = nose_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping


def _build_upper_eyelid():
    top_indices = np.arange(0, 7)
    middle_indices = np.arange(12, 17)
    upper_eyelid_indices = np.hstack((top_indices, middle_indices))

    upper_eyelid_connectivity = list(zip(top_indices, top_indices[1:]))
    upper_eyelid_connectivity += [(0, 12)]
    upper_eyelid_connectivity += list(zip(middle_indices, middle_indices[1:]))
    upper_eyelid_connectivity += [(16, 6)]

    return upper_eyelid_indices, upper_eyelid_connectivity


@labeller_func(group_label="eye_ibug_open_38")
def eye_ibug_open_38_to_eye_ibug_open_38(pcloud):
    r"""
    Apply the IBUG 38-point open eye semantic labels.

    The semantic labels applied are as follows:

      - upper_eyelid
      - lower_eyelid
      - iris
      - pupil
      - sclera
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 38
    validate_input(pcloud, n_expected_points)

    upper_el_indices, upper_el_connectivity = _build_upper_eyelid()

    iris_range = (22, 30)
    pupil_range = (30, 38)
    sclera_top = np.arange(12, 17)
    sclera_bottom = np.arange(17, 22)
    sclera_indices = np.hstack((0, sclera_top, 6, sclera_bottom))
    lower_el_top = np.arange(17, 22)
    lower_el_bottom = np.arange(7, 12)
    lower_el_indices = np.hstack((6, lower_el_top, 0, lower_el_bottom))

    iris_connectivity = connectivity_from_range(iris_range, close_loop=True)
    pupil_connectivity = connectivity_from_range(pupil_range, close_loop=True)

    sclera_connectivity = list(zip(sclera_top, sclera_top[1:]))
    sclera_connectivity += [(0, 21)]
    sclera_connectivity += list(zip(sclera_bottom, sclera_bottom[1:]))
    sclera_connectivity += [(6, 17)]

    lower_el_connectivity = list(zip(lower_el_top, lower_el_top[1:]))
    lower_el_connectivity += [(6, 7)]
    lower_el_connectivity += list(zip(lower_el_bottom, lower_el_bottom[1:]))
    lower_el_connectivity += [(11, 0)]

    all_connectivity = np.asarray(
        upper_el_connectivity
        + lower_el_connectivity
        + iris_connectivity.tolist()
        + pupil_connectivity.tolist()
        + sclera_connectivity
    )

    mapping = OrderedDict()
    mapping["upper_eyelid"] = upper_el_indices
    mapping["lower_eyelid"] = lower_el_indices
    mapping["pupil"] = np.arange(*pupil_range)
    mapping["iris"] = np.arange(*iris_range)
    mapping["sclera"] = sclera_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="eye_ibug_close_17")
def eye_ibug_close_17_to_eye_ibug_close_17(pcloud):
    r"""
    Apply the IBUG 17-point close eye semantic labels.

    The semantic labels applied are as follows:

      - upper_eyelid
      - lower_eyelid
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 17
    validate_input(pcloud, n_expected_points)

    upper_indices, upper_connectivity = _build_upper_eyelid()

    middle_indices = np.arange(12, 17)
    bottom_indices = np.arange(6, 12)
    lower_indices = np.hstack((bottom_indices, 0, middle_indices))
    lower_connectivity = list(zip(bottom_indices, bottom_indices[1:]))
    lower_connectivity += [(0, 12)]
    lower_connectivity += list(zip(middle_indices, middle_indices[1:]))
    lower_connectivity += [(11, 0)]

    all_connectivity = np.asarray(upper_connectivity + lower_connectivity)

    mapping = OrderedDict()
    mapping["upper_eyelid"] = upper_indices
    mapping["lower_eyelid"] = lower_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping
    )

    return new_pcloud, mapping


@labeller_func(group_label="eye_ibug_open_38_trimesh")
def eye_ibug_open_38_to_eye_ibug_open_38_trimesh(pcloud):
    r"""
    Apply the IBUG 38-point open eye semantic labels, with trimesh connectivity.

    The semantic labels applied are as follows:

      - tri
    """
    from menpo.shape import TriMesh

    n_expected_points = 38
    validate_input(pcloud, n_expected_points)

    tri_list = np.array(
        [
            [29, 36, 28],
            [22, 13, 23],
            [12, 1, 2],
            [29, 30, 37],
            [13, 3, 14],
            [13, 12, 2],
            [19, 8, 9],
            [25, 33, 24],
            [36, 37, 33],
            [24, 32, 31],
            [33, 37, 31],
            [35, 34, 27],
            [35, 36, 33],
            [3, 13, 2],
            [14, 24, 23],
            [33, 32, 24],
            [15, 25, 14],
            [25, 26, 34],
            [22, 30, 29],
            [31, 37, 30],
            [24, 31, 23],
            [32, 33, 31],
            [22, 12, 13],
            [0, 1, 12],
            [14, 23, 13],
            [31, 30, 23],
            [28, 19, 20],
            [21, 11, 0],
            [12, 21, 0],
            [20, 11, 21],
            [20, 10, 11],
            [21, 29, 20],
            [21, 12, 22],
            [30, 22, 23],
            [29, 21, 22],
            [27, 19, 28],
            [29, 37, 36],
            [29, 28, 20],
            [36, 35, 28],
            [20, 19, 10],
            [10, 19, 9],
            [28, 35, 27],
            [19, 19, 8],
            [17, 16, 6],
            [18, 7, 8],
            [25, 34, 33],
            [18, 27, 17],
            [18, 19, 27],
            [18, 17, 7],
            [27, 26, 17],
            [17, 6, 7],
            [14, 25, 24],
            [34, 35, 33],
            [17, 26, 16],
            [27, 34, 26],
            [3, 15, 14],
            [15, 26, 25],
            [4, 15, 3],
            [16, 26, 15],
            [16, 4, 5],
            [16, 15, 4],
            [16, 5, 6],
            [8, 18, 19],
        ]
    )

    new_pcloud = TriMesh(pcloud.points, trilist=tri_list, copy=False)

    mapping = OrderedDict()
    mapping["tri"] = np.arange(new_pcloud.n_points)

    return new_pcloud, mapping


@labeller_func(group_label="eye_ibug_close_17_trimesh")
def eye_ibug_close_17_to_eye_ibug_close_17_trimesh(pcloud):
    r"""
    Apply the IBUG 17-point close eye semantic labels, with trimesh
    connectivity.

    The semantic labels applied are as follows:

      - tri
    """
    from menpo.shape import TriMesh

    n_expected_points = 17
    validate_input(pcloud, n_expected_points)

    tri_list = np.array(
        [
            [10, 11, 13],
            [3, 13, 2],
            [4, 14, 3],
            [15, 5, 16],
            [12, 11, 0],
            [13, 14, 10],
            [13, 12, 2],
            [14, 13, 3],
            [0, 1, 12],
            [2, 12, 1],
            [13, 11, 12],
            [9, 10, 14],
            [15, 9, 14],
            [7, 8, 15],
            [5, 6, 16],
            [15, 14, 4],
            [7, 15, 16],
            [8, 9, 15],
            [15, 4, 5],
            [16, 6, 7],
        ]
    )

    new_pcloud = TriMesh(pcloud.points, trilist=tri_list, copy=False)

    mapping = OrderedDict()
    mapping["tri"] = np.arange(new_pcloud.n_points)

    return new_pcloud, mapping


@labeller_func(group_label="tongue_ibug_19")
def tongue_ibug_19_to_tongue_ibug_19(pcloud):
    r"""
    Apply the IBUG 19-point tongue semantic labels.

    The semantic labels applied are as follows:

      - outline
      - bisector
    """
    n_expected_points = 19
    validate_input(pcloud, n_expected_points)

    labels = OrderedDict([("outline", (0, 13, False)), ("bisector", (13, 19, False))])
    return pcloud_and_lgroup_from_ranges(pcloud, labels)



import os
import re
import subprocess as sp
import warnings
from pathlib import Path

import numpy as np

from menpo.base import LazyList
from menpo.image import Image
from menpo.image.base import normalize_pixels_range
from ..utils import _call_subprocess

_FFMPEG_CMD = lambda: str(Path(os.environ.get("MENPO_FFMPEG_CMD", "ffmpeg")))
_FFPROBE_CMD = lambda: str(Path(os.environ.get("MENPO_FFPROBE_CMD", "ffprobe")))


def ffmpeg_importer(filepath, normalize=True, exact_frame_count=True, **kwargs):
    r"""
    Imports videos by streaming frames from a pipe using FFMPEG. Returns a
    :map:`LazyList` that gives lazy access to the video on a per-frame basis.

    There are two important environment variables that can be set to alter
    the behaviour of this function:

        ================== ======================================
        ENV Variable       Definition
        ================== ======================================
        MENPO_FFMPEG_CMD   The path to the 'ffmpeg' executable.
        MENPO_FFPROBE_CMD  The path to the 'ffprobe' executable.
        ================== ======================================

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the video.
    normalize : `bool`, optional
        If ``True``, normalize between 0.0 and 1.0 and convert to float. If
        ``False`` just return whatever ffmpeg imports.
    exact_frame_count: `bool`, optional
        If ``True``, the import fails if ffprobe is not available
        (reading from ffmpeg's output returns inexact frame count)
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    image : :map:`LazyList`
        A :map:`LazyList` containing :map:`Image` or subclasses per frame
        of the video.
    """
    reader = FFMpegVideoReader(
        filepath, normalize=normalize, exact_frame_count=exact_frame_count
    )
    ll = LazyList.init_from_index_callable(
        lambda x: Image.init_from_channels_at_back(reader[x]), len(reader)
    )
    ll.fps = reader.fps

    return ll


def ffmpeg_types():
    r"""The supported FFMPEG types.

    Returns
    -------
    supported_types : `dict`
        A dictionary of extensions supported by the FFMPEG importer.
    """
    # Need to check this
    ffmpeg_exts = [".avi", ".mp4", ".mpg", ".mpeg", ".wmv", ".mov", ".mkv"]
    return {ext: ffmpeg_importer for ext in ffmpeg_exts}


class FFMpegVideoReader(object):
    """
    Read a video using ffmpeg and handle state to allow seeking.

    Parameters
    ----------
    filepath : `Path`
        Absolute path to the video
    normalize : `bool`, optional
        If ``True``, the resulting range of the pixels of the returned
        frames is normalized.
    exact_frame_count : `bool`, optional
        If True, the import fails if ffmprobe is not available
        (reading from ffmpeg's output returns inexact frame count)
    """

    def __init__(self, filepath, normalize=False, exact_frame_count=True):
        self.filepath = filepath
        self.normalize = normalize
        self.exact_frame_count = exact_frame_count
        self._pipe = None
        if self.exact_frame_count:
            try:
                infos = video_infos_ffprobe(self.filepath)
            except:
                raise ValueError(
                    "ffprobe not available, unable to get exact "
                    "frame count. If you want to use an "
                    "approximate frame number, set exact_frame_"
                    "count to False and proceed at your own "
                    "risk."
                )
        else:
            infos = video_infos_ffmpeg(self.filepath)
        self.duration = infos["duration"]
        self.width = infos["width"]
        self.height = infos["height"]
        self.n_frames = infos["n_frames"]
        self.fps = infos["fps"]
        # contains the index of the last read frame
        # the index is updated in _open_pipe, _read_one_frame and _trash_frames
        self.index = -1

    def _shutdown_pipe(self):
        if self._pipe is not None:
            if self._pipe.stdout:
                self._pipe.stdout.close()
            if self._pipe.stderr:
                self._pipe.stderr.close()
            if self._pipe.stdin:
                self._pipe.stdin.close()
        self._pipe = None

    def __del__(self):
        r"""
        Close the pipe if open.
        """
        self._shutdown_pipe()

    def __len__(self):
        return self.n_frames

    def _open_pipe(self, frame=None):
        r"""
        Open a pipe at the time just before the specified frame

        Parameters
        ----------
        frame : `int`, optional
            If ``None``, pipe opened from the beginning of the video
            otherwise, pipe opened at the time corresponding to that frame

        Note
        ----
        Since v.2.1 of ffmpeg, this is frame-accurate
        """
        if frame is not None and frame > 0:
            time = str(frame / float(self.fps))
            command = [
                _FFMPEG_CMD(),
                "-ss",
                time,
                "-i",
                str(self.filepath),
                "-f",
                "image2pipe",
                "-pix_fmt",
                "rgb24",
                "-vcodec",
                "rawvideo",
                "-",
            ]
        else:
            command = [
                _FFMPEG_CMD(),
                "-i",
                str(self.filepath),
                "-f",
                "image2pipe",
                "-pix_fmt",
                "rgb24",
                "-vcodec",
                "rawvideo",
                "-",
            ]
            frame = 0

        self._shutdown_pipe()
        self._pipe = sp.Popen(
            command,
            stdout=sp.PIPE,
            stdin=sp.DEVNULL,
            stderr=sp.DEVNULL,
            bufsize=10 ** 8,
        )  # Is this buffer the correct size?
        # We have not yet read the specified frame
        self.index = frame - 1

    def __iter__(self):
        r"""
        Iterate through all frames of the video in order

        Only opens the pipe once at the beginning
        """
        self.index = 0
        for index in range(self.n_frames):
            yield self[index]

    def __getitem__(self, index):
        r"""
        Get a specific frame from the video
        """
        # If the user is reading consecutive frames, or a frame later in the
        # video, do not reopen a pipe
        if self._pipe is None or self._pipe.poll() is not None or index <= self.index:
            self._open_pipe(frame=index)
        else:
            to_trash = index - self.index - 1
            if to_trash > 0:
                self._trash_frames(to_trash)

        return self._read_one_frame()

    def _trash_frames(self, n_frames):
        r"""
        Reads and trashes the data corresponding to ``n_frames``
        """
        _ = self._pipe.stdout.read(self.height * self.width * 3 * n_frames)
        self._pipe.stdout.flush()
        self.index += n_frames

    def _read_one_frame(self):
        r"""
        Reads one frame from the opened ``self._pipe`` and converts it to
        a numpy array

        Returns
        -------
        image : :map:`Image`
            Image of shape ``(self.height, self.width, 3)``
        """
        raw_data = self._pipe.stdout.read(self.height * self.width * 3)
        frame = np.frombuffer(raw_data, dtype=np.uint8)
        frame = frame.reshape((self.height, self.width, 3))
        self._pipe.stdout.flush()
        self.index += 1

        if self.normalize:
            frame = normalize_pixels_range(frame)

        return frame


def video_infos_ffmpeg(filepath):
    r"""
    Parses the information from a video using ffmpeg.
    Uses subprocess to get the information through a pipe.

    Parameters
    ----------
    filepath : `Path`
        absolute path to the video file which information to extract

    Returns
    -------
    infos : `dict`
        keys are width, height (size of the frames)
        duration (duration of the video in seconds)
        n_frames
    """
    warnings.warn(
        "Estimating number of frames using ffmpeg duration which "
        "may be inaccurate for certain types of encodings. Try "
        "setting the MENPO_FFPROBE_CMD environment variable to "
        "define the path to ffprobe."
    )

    # Read information using ffmpeg - the call below intentionally causes
    # an error about no output from FFMPEG in order to terminate faster - hence
    # reading the output from stderr.
    command = [_FFMPEG_CMD(), "-i", str(filepath), "-"]
    with _call_subprocess(sp.Popen(command, stdout=sp.DEVNULL, stderr=sp.PIPE)) as pipe:
        raw_infos = pipe.stderr.read().decode()

    # Note: we use '\d+\.?\d*' so we can match both int and float for the fps
    video_info = re.search(
        r"Video:.*(?P<width> \d+)x(?P<height>\d+).*(?P<fps> \d+\.?\d*) fps",
        raw_infos,
        re.DOTALL,
    ).groupdict()

    # Some videos don't have a valid duration
    time = re.search(
        r"Duration:\s(?P<hours>\d+?):(?P<minutes>\d+?):(?P<seconds>\d+\.\d+?),",
        raw_infos,
        re.DOTALL,
    )
    if time is None:
        raise ValueError(
            "Unable to determine duration for video - please "
            "install and use ffprobe for accurate frame count "
            "computation."
        )

    # Get the duration in seconds and convert size to ints
    time = time.groupdict()
    hours = float(time["hours"])
    minutes = float(time["minutes"])
    seconds = float(time["seconds"])
    duration = 60 * 60 * hours + 60 * minutes + seconds

    fps = round(float(video_info["fps"]))
    n_frames = round(duration * fps)
    width = int(video_info["width"])
    height = int(video_info["height"])

    # Create the resulting dictionary
    infos = {
        "duration": duration,
        "width": width,
        "height": height,
        "n_frames": n_frames,
        "fps": fps,
    }

    return infos


def video_infos_ffprobe(filepath):
    """
    Parses the information from a video using ffprobe
    Uses subprocess to get the information through a pipe

    Parameters
    ----------
    filepath : `Path`
        Absolute path to the video file which information to extract

    Returns
    -------
    infos : `dict`
        keys are width, height (size of the frames)
        duration (duration of the video in seconds)
        n_frames
    """
    expected_keys = {"width", "height", "avg_frame_rate", "duration", "nb_read_frames"}

    p = sp.Popen(
        [
            _FFPROBE_CMD(),
            "-v",
            "quiet",  # Quiet output
            "-count_frames",  # Count the number of complete frames
            "-select_streams",
            "v:0",  # Only show the first stream
            "-show_entries",  # Show the entries below
            "stream=height,width,nb_read_frames,duration,avg_frame_rate",
            "-of",
            "default=noprint_wrappers=1",  # Output format is just to print key=value pairs
            str(filepath),
        ],
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
    )
    with _call_subprocess(p) as pipe:
        stdout_output = pipe.stdout.readlines()
    del p

    # Split the key=value pairs on the '='
    kv_dict = dict([v.decode().strip().split("=") for v in stdout_output])
    found_keys = set(kv_dict.keys())

    if found_keys ^ expected_keys:
        raise ValueError(
            "Not all of the expected values were returned. "
            "Expected {} but found {}.".format(expected_keys, found_keys)
        )

    kv_dict["n_frames"] = int(kv_dict.pop("nb_read_frames"))
    kv_dict["width"] = int(kv_dict["width"])
    kv_dict["height"] = int(kv_dict["height"])
    # Some videos may not have a valid duration
    if kv_dict["duration"] == "N/A":
        kv_dict["duration"] = None
    else:
        kv_dict["duration"] = float(kv_dict["duration"])
    fps = kv_dict.pop("avg_frame_rate").split("/")
    try:
        kv_dict["fps"] = float(fps[0]) / float(fps[1])
    except ZeroDivisionError:
        kv_dict["fps"] = None

    return kv_dict
